# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Deeper testing of some DateParser internals.
"""

import unittest

from ...utils.grampslocale import GrampsLocale
from ...lib.date import Date


class DateDisplayTest(unittest.TestCase):
    def setUp(self):
        from .._datedisplay import DateDisplayEn

        self.display = DateDisplayEn()
        self.display_RU = GrampsLocale(lang="ru").date_displayer

    def assert_map_key_val(self, m, k, v):
        try:
            self.assertEqual(m[k], v)
        except KeyError:
            self.assertTrue(False, list(m.items()))


class DateDisplayCalendarTest(DateDisplayTest):
    def test_calendar_gregorian_is_empty(self):
        self.assert_map_key_val(self.display.calendar, Date.CAL_GREGORIAN, "")

    def test_calendar_julian_RU(self):
        self.assert_map_key_val(self.display_RU.calendar, Date.CAL_JULIAN, "Юлианский")


# This class tests common functionality in DateDisplay as applied to RU,
# and so it is coupled to translated strings and inflection names
# extracted by lexgettext from ru.po
class DateDisplayInflectionsTestRU(DateDisplayTest):
    def setUp(self):
        DateDisplayTest.setUp(self)
        self.dd = self.display = self.display_RU
        self.months = self.dd._ds.long_months
        # TODO hardwired magic numbers! Bad API smell.
        self.dd.set_format(4)  # day month_name year
        self.may = self.months[5]

    def assertInflectionInDate(self, inflection, date, month=None):
        if month is None:
            month = date.get_month()
        month_lexeme = self.months[month]
        self.assertIn(month_lexeme.forms[inflection], self.dd.display(date))

    def test_month_only_date_nominative_quality_none(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_quality(Date.QUAL_NONE)
        self.assertInflectionInDate("И", d1945may)

    def test_month_only_date_nominative_quality_estimated(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_quality(Date.QUAL_ESTIMATED)
        self.assertInflectionInDate("Т", d1945may)

    def test_month_only_date_nominative_quality_calculated(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_quality(Date.QUAL_CALCULATED)
        self.assertInflectionInDate("И", d1945may)

    def test_day_month_date_genitive(self):
        d1945may9 = Date(1945, 5, 9)
        self.assertInflectionInDate("Р", d1945may9)

    def test_day_month_date_genitiive_quality_estimated(self):
        d1945may9 = Date(1945, 5, 9)
        d1945may9.set_quality(Date.QUAL_ESTIMATED)
        self.assertInflectionInDate("Р", d1945may9)

    def test_before_month_only_date_genitive(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_modifier(Date.MOD_BEFORE)
        # TODO hardwired magic numbers! Bad API smell.
        for inflecting_format in (3, 4):
            self.dd.set_format(inflecting_format)
            # this depends on the fact that in Russian the short and long forms for May
            # will be the same!
            self.assertIn("до мая", self.dd.display(d1945may))

    def test_after_month_only_date_genitive(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_modifier(Date.MOD_AFTER)
        # TODO hardwired magic numbers! Bad API smell.
        for inflecting_format in (3, 4):
            self.dd.set_format(inflecting_format)
            # this depends on the fact that in Russian the short and long forms for May
            # will be the same!
            self.assertIn("после мая", self.dd.display(d1945may))

    def test_about_month_only_date_genitive(self):
        d1945may = Date(1945, 5, 0)
        d1945may.set_modifier(Date.MOD_ABOUT)
        # TODO hardwired magic numbers! Bad API smell.
        for inflecting_format in (3, 4):
            self.dd.set_format(inflecting_format)
            # this depends on the fact that in Russian the short and long forms for May
            # will be the same!
            self.assertIn("около мая", self.dd.display(d1945may))

    def test_between_month_only_dates_ablative(self):
        b1945may_1946may = Date()
        b1945may_1946may.set(
            modifier=Date.MOD_RANGE, value=(0, 5, 1945, False, 0, 5, 1946, False)
        )
        # TODO hardwired magic numbers! Bad API smell.
        for inflecting_format in (3, 4):
            self.dd.set_format(inflecting_format)
            # this depends on the fact that in Russian the short and long forms for May
            # will be the same!
            self.assertIn("между маем", self.dd.display(b1945may_1946may))
            self.assertIn("и маем", self.dd.display(b1945may_1946may))

    def test_month_only_date_span_from_genitive_to_accusative(self):
        f1945may_t1946may = Date()
        f1945may_t1946may.set(
            modifier=Date.MOD_SPAN, value=(0, 5, 1945, False, 0, 5, 1946, False)
        )
        # TODO hardwired magic numbers! Bad API smell.
        for inflecting_format in (3, 4):
            self.dd.set_format(inflecting_format)
            # this depends on the fact that in Russian the short and long forms for May
            # will be the same!
            self.assertIn("с мая", self.dd.display(f1945may_t1946may))
            self.assertIn("по май", self.dd.display(f1945may_t1946may))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2024  Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import logging

LOG = logging.getLogger(".filter.optimizer")


def intersection(sets):
    if sets:
        result = sets[0]
        for s in sets[1:]:
            result = result.intersection(s)
        return result
    else:
        return set()


def union(sets):
    if sets:
        result = sets[0]
        for s in sets[1:]:
            result = result.union(s)
        return result
    else:
        return set()


class Optimizer:
    """
    Optimizer to use the filter's pre-selected selected_handles
    to include or exclude.
    """

    def __init__(self, filter):
        """
        Initialize the collection of selected_handles in the filter list.
        """
        self.all_selected_handles = []
        self.walk_filters(filter, False, self.all_selected_handles)

    def walk_filters(self, filter, parent_invert, result):
        """
        Recursively walk all of the filters/rules and get
        rules with selected_handles
        """
        current_invert = parent_invert if not filter.invert else not parent_invert
        LOG.debug(
            "walking, filter: %s, invert=%s, parent_invert=%s",
            filter,
            filter.invert,
            parent_invert,
        )
        rules_with_selected_handles = []
        for item in filter.flist:
            if hasattr(item, "find_filter"):
                rule_filter = item.find_filter()
                if rule_filter is not None:
                    self.walk_filters(
                        rule_filter,
                        current_invert,
                        result,
                    )
            elif hasattr(item, "selected_handles"):
                rules_with_selected_handles.append(set(item.selected_handles))
        if rules_with_selected_handles:
            LOG.debug(
                "filter %s: parent_invert=%s, invert=%s, op=%s, number of rules with selected_handles=%s",
                filter,
                parent_invert,
                filter.invert,
                filter.logical_op,
                len(rules_with_selected_handles),
            )
            result.append(
                (
                    current_invert,
                    filter.logical_op,
                    rules_with_selected_handles,
                )
            )

    def get_handles(self):
        """
        Returns handles_in, and handles_out.

        `handles_in` is either None, or a set of handles to include.
        if it is None, then there is no evidence to only include
        particular handles. If it is a set, then those in the set
        are a superset of the items that will match.

        `handles_out` is a set. If any handle is in the set, it will
        not be included in the final results.

        The handles_in are selected if all of the rules are connected with
        "and" and not inverted. The handles_out are selected if all of the
        rules are connected with "and" and inverted.
        """
        handles_in = None
        handles_out = set()
        # Get all positive non-inverted selected_handles
        for inverted, logical_op, selected_handles in self.all_selected_handles:
            if logical_op == "and" and not inverted:
                LOG.debug("optimizer positive match!")
                if handles_in is None:
                    handles_in = intersection(selected_handles)
                else:
                    handles_in = intersection([handles_in] + selected_handles)

        # Get all inverted selected_handles:
        for inverted, logical_op, selected_handles in self.all_selected_handles:
            if logical_op == "and" and inverted:
                LOG.debug("optimizer inverted match!")
                handles_out = union([handles_out] + selected_handles)

        if handles_in is not None:
            handles_in = handles_in - handles_out

        LOG.debug("optimizer handles_in: %s", len(handles_in) if handles_in else 0)
        return handles_in, handles_out



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen.filters.rules/Place/_InLatLonNeighborhood.py


# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....utils.place import conv_lat_lon


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Union
from ....lib import Place
from ....db import Database


# -------------------------------------------------------------------------
#
# InLatLonNeighborhood
#
# -------------------------------------------------------------------------
class InLatLonNeighborhood(Rule):
    """Rule that checks if a place is in the neighborhood of a given
    Latitude or Longitude or"""

    labels = [
        _("Latitude:"),
        _("Longitude:"),
        _("Rectangle height:"),
        _("Rectangle width:"),
    ]
    name = _("Places in neighborhood of given position")
    description = _(
        "Matches places with latitude or longitude positioned in "
        "a rectangle of given height and width (in degrees), and "
        "with middlepoint the given latitude and longitude."
    )
    category = _("Position filters")

    def prepare(self, db: Database, user):
        self.halfheight: Union[float, None] = None
        self.halfwidth: Union[float, None] = None
        if self.list[0]:
            try:
                self.halfheight = float(self.list[2]) / 2.0
            except ValueError:
                self.halfheight = None
            if self.halfheight is not None and self.halfheight <= 0.0:
                self.halfheight = None
        else:
            self.halfheight = -1
            # give dummy value
            self.list[0] = "0.0"

        if self.list[1]:
            try:
                self.halfwidth = float(self.list[3]) / 2.0
            except ValueError:
                self.halfwidth = None
            if self.halfwidth is not None and self.halfwidth <= 0.0:
                self.halfwidth = None
        else:
            self.halfwidth = -1
            # give dummy value
            self.list[1] = "0.0"

        # we allow a band instead of a triangle
        results = conv_lat_lon(self.list[0], self.list[1], "D.D8")  # type: ignore
        self.lat, self.lon = results if isinstance(results, tuple) else (None, None)  # type: ignore
        if self.lat is not None and self.lon is not None:
            self.lat = float(self.lat)
            self.lon = float(self.lon)
        else:
            self.lat = None
            self.lon = None

        # we define the two squares we must look in
        #    can be 0, so check on None
        if (
            self.lat is not None
            and self.halfheight is not None
            and self.halfheight != -1
        ):
            self.S = self.lat + self.halfheight
            if self.S > 90.0:
                self.S = 90.0
            self.N = self.lat - self.halfheight
            if self.N < -90.0:
                self.N = -90.0
        self.doublesquares = False
        if self.lon is not None and self.halfwidth is not None and self.halfwidth != -1:
            if self.halfwidth >= 180.0:
                # the entire longitude is allowed, reset values
                self.lon = 0.0
                self.E = 180.0
                self.W = -180.0
            else:
                self.E = self.lon + self.halfwidth
                self.W = self.lon - self.halfwidth
                if self.E > 180.0:
                    # we need to check in two squares:
                    self.doublesquares = True
                    self.E2 = self.E - 360.0
                    self.W2 = -180.0
                    self.E = 180
                if self.W < -180.0:
                    # we need to check in two squares:
                    self.doublesquares = True
                    self.W2 = self.W + 360.0
                    self.E2 = 180.0
                    self.W = -180

    def apply_to_one(self, db: Database, place: Place) -> bool:
        if self.halfheight == -1 and self.halfwidth == -1:
            return False

        # when given, must be valid
        if self.lat is None or self.lon is None:
            return False

        # if height/width given, they must be valid
        if self.halfheight is None or self.halfwidth is None:
            return False

        # now we know at least one is given in the filter and is valid

        # the place we look at must have lat AND lon entered
        if not (place.lat.strip() and place.long.strip()):
            return False

        results = conv_lat_lon(place.lat, place.long, "D.D8")  # type: ignore
        latpl, lonpl = results if isinstance(results, tuple) else (None, None)  # type: ignore
        if latpl and lonpl:
            latpl = float(latpl)  # type: ignore
            lonpl = float(lonpl)  # type: ignore
            if self.halfheight != -1:
                # check lat
                if latpl < self.N or latpl > self.S:  # type: ignore
                    return False

            if self.halfwidth != -1:
                # check lon: more difficult, we may cross the 180/-180 boundary
                # and must keep counting
                if self.doublesquares:
                    # two squares to look in :
                    left = lonpl < self.W or lonpl > self.E  # type: ignore
                    right = lonpl < self.W2 or lonpl > self.E2  # type: ignore
                    if left and right:  # type: ignore
                        return False
                elif lonpl < self.W or lonpl > self.E:  # type: ignore
                    return False

            return True

        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....utils.db import for_each_ancestor
from ._hascommonancestorwith import HasCommonAncestorWith
from ._matchesfilter import MatchesFilter


# -------------------------------------------------------------------------
#
# HasCommonAncestorWithFilterMatch
#
# -------------------------------------------------------------------------
class HasCommonAncestorWithFilterMatch(HasCommonAncestorWith):
    """Rule that checks for a person that has a common ancestor with
    someone matching a filter"""

    labels = [_("Filter name:")]
    name = _("People with a common ancestor with <filter> match")
    description = _(
        "Matches people that have a common ancestor " "with anybody matched by a filter"
    )
    category = _("Ancestral filters")

    def __init__(self, list, use_regex=False, use_case=False):
        HasCommonAncestorWith.__init__(self, list, use_regex, use_case)
        self.ancestor_cache = {}

    def prepare(self, db, user):
        self.db = db
        # For each(!) person we keep track of who their ancestors
        # are, in a set(). So we only have to compute a person's
        # ancestor list once.
        # Start with filling the cache for root person (gramps_id in self.list[0])
        self.ancestor_cache = {}
        self.with_people = []
        self.filt = MatchesFilter(self.list)
        self.filt.requestprepare(db, user)
        if user:
            user.begin_progress(
                self.category,
                _("Retrieving all sub-filter matches"),
                db.get_number_of_people(),
            )
        for person in db.iter_people():
            if user:
                user.step_progress()
            if person and self.filt.apply_to_one(db, person):
                # store all people in the filter so as to compare later
                self.with_people.append(person.handle)
                # fill list of ancestor of person if not present yet
                if person.handle not in self.ancestor_cache:
                    self.add_ancs(db, person)
        if user:
            user.end_progress()

    def reset(self):
        self.filt.requestreset()
        self.ancestor_cache = {}



##
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007       Brian G. Matherly
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013,2015  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide utilities for printing endnotes in text reports.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..docgen import FontStyle, ParagraphStyle, FONT_SANS_SERIF
from ...lib import NoteType, Citation
from ...utils.string import conf_strings
from ...config import config
from ..utils import get_cite


def add_endnote_styles(style_sheet):
    """
    Add paragraph styles to a style sheet to be used for displaying endnotes.

    :param style_sheet: Style sheet
    :type style_sheet: :class:`.StyleSheet`
    """
    font = FontStyle()
    font.set(face=FONT_SANS_SERIF, size=14, italic=1)
    para = ParagraphStyle()
    para.set_font(font)
    para.set_header_level(2)
    para.set_top_margin(0.2)
    para.set_bottom_margin(0.2)
    para.set_description(_("The style used for the generation header."))
    style_sheet.add_paragraph_style("Endnotes-Header", para)

    para = ParagraphStyle()
    para.set(first_indent=-0.75, lmargin=0.75)
    para.set_top_margin(0.2)
    para.set_bottom_margin(0.0)
    para.set_description(_("The basic style used for " "the endnotes source display."))
    style_sheet.add_paragraph_style("Endnotes-Source", para)

    para = ParagraphStyle()
    para.set(lmargin=0.75)
    para.set_top_margin(0.2)
    para.set_bottom_margin(0.0)
    para.set_description(_("The basic style used for " "the endnotes notes display."))
    style_sheet.add_paragraph_style("Endnotes-Source-Notes", para)

    para = ParagraphStyle()
    para.set(first_indent=-0.9, lmargin=1.9)
    para.set_top_margin(0.2)
    para.set_bottom_margin(0.0)
    para.set_description(
        _("The basic style used for " "the endnotes reference display.")
    )
    style_sheet.add_paragraph_style("Endnotes-Ref", para)

    para = ParagraphStyle()
    para.set(lmargin=1.9)
    para.set_top_margin(0.2)
    para.set_bottom_margin(0.0)
    para.set_description(
        _("The basic style used for " "the endnotes reference notes display.")
    )
    style_sheet.add_paragraph_style("Endnotes-Ref-Notes", para)


def cite_source(bibliography, database, obj, elocale=glocale):
    """
    Cite any sources for the object and add them to the bibliography.

    :param bibliography: The bibliography to contain the citations.
    :type bibliography: :class:`.Bibliography`
    :param obj: An object with source references.
    :type obj: :class:`~.citationbase.CitationBase`
    :param elocale: allow deferred translation of dates and strings
    :type elocale: a :class:`.GrampsLocale` instance
    """

    trans_text = elocale.translation.gettext
    # trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)

    txt = ""
    slist = obj.get_citation_list()
    if slist:
        first = True
        for ref in slist:
            if not first:
                # Translators: needed for Arabic, ignore otherwise
                txt += trans_text(", ")
            first = False
            citation = database.get_citation_from_handle(ref)
            (cindex, key) = bibliography.add_reference(citation)
            txt += "%d" % (cindex + 1)
            if key is not None:
                txt += key
    return txt


def write_endnotes(
    bibliography, database, doc, printnotes=False, links=False, elocale=glocale
):
    """
    Write all the entries in the bibliography as endnotes.

    If elocale is passed in (a :class:`.GrampsLocale`), then (insofar as
    possible) the translated values will be returned instead.

    :param bibliography: The bibliography that contains the citations.
    :type bibliography: :class:`.Bibliography`
    :param database: The database that the sources come from.
    :type database: DbBase
    :param doc: The document to write the endnotes into.
    :type doc: :class:`~.docgen.TextDoc`
    :param printnotes: Indicate if the notes attached to a source must be
            written too.
    :type printnotes: bool
    :param links: Indicate if URL links should be makde 'clickable'.
    :type links: bool
    :param elocale: allow deferred translation of dates and strings
    :type elocale: a :class:`.GrampsLocale` instance
    """
    if bibliography.get_citation_count() == 0:
        return

    trans_text = elocale.translation.gettext
    # trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)

    doc.start_paragraph("Endnotes-Header")
    doc.write_text(trans_text("Endnotes"))
    doc.end_paragraph()

    cindex = 0
    cite = get_cite()
    for citation in cite.format(bibliography, database, elocale):
        cindex += 1

        doc.start_paragraph("Endnotes-Source", "%d." % cindex)
        source, text = citation[0]
        doc.write_text(text, links=links)
        doc.end_paragraph()

        if printnotes:
            _print_notes(source, database, doc, "Endnotes-Source-Notes", links)

        for ref, key, text in citation[1]:
            # Translators: needed for French, ignore otherwise
            doc.start_paragraph("Endnotes-Ref", trans_text("%s:") % key)
            doc.write_text(text, links=links)
            doc.end_paragraph()

            if printnotes:
                _print_notes(ref, database, doc, "Endnotes-Ref-Notes", links)


def _print_notes(obj, db, doc, style, links):
    note_list = obj.get_note_list()
    ind = 1
    for notehandle in note_list:
        note = db.get_note_from_handle(notehandle)
        contains_html = note.get_type() == NoteType.HTML_CODE
        doc.write_styled_note(
            note.get_styledtext(),
            note.get_format(),
            style,
            contains_html=contains_html,
            links=links,
        )
        ind += 1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009   Stephane Charette
# Copyright (C) 2016-       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from collections import OrderedDict

"Find possible loop in a people descendance"

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gui.plug import tool
from gramps.gui.editors import EditFamily
from gramps.gen.errors import WindowActiveError
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.utils import ProgressMeter
from gramps.gui.display import display_help
from gramps.gui.glade import Glade
from gramps.gen.display.name import displayer as _nd
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Find_database_loop", "manual")


# ------------------------------------------------------------------------
#
# FindLoop class
#
# ------------------------------------------------------------------------
class FindLoop(ManagedWindow):
    """
    Find loops in the family tree.
    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate

        self.title = _("Find database loop")
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate
        self.uistate = uistate
        # self.db = CacheProxyDb(dbstate.db)
        self.db = dbstate.db

        top_dialog = Glade()

        top_dialog.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_event": self.close,
            }
        )

        window = top_dialog.toplevel
        title = top_dialog.get_object("title")
        self.set_window(window, title, self.title)

        # start the progress indicator
        self.progress = ProgressMeter(self.title, _("Starting"), parent=uistate.window)
        self.progress.set_pass(
            _("Looking for possible loop for each person"),
            self.db.get_number_of_people(),
        )

        self.model = Gtk.ListStore(
            GObject.TYPE_STRING,  # 0==father id
            GObject.TYPE_STRING,  # 1==father
            GObject.TYPE_STRING,  # 2==son id
            GObject.TYPE_STRING,  # 3==son
            GObject.TYPE_STRING,  # 4==family gid
            GObject.TYPE_STRING,
        )  # 5==loop number
        self.model.set_sort_column_id(Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 0)

        self.treeview = top_dialog.get_object("treeview")
        self.treeview.set_model(self.model)
        col0 = Gtk.TreeViewColumn("", Gtk.CellRendererText(), text=5)
        col1 = Gtk.TreeViewColumn(_("Gramps ID"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("Parent"), Gtk.CellRendererText(), text=1)
        col3 = Gtk.TreeViewColumn(_("Gramps ID"), Gtk.CellRendererText(), text=2)
        col4 = Gtk.TreeViewColumn(_("Child"), Gtk.CellRendererText(), text=3)
        col5 = Gtk.TreeViewColumn(_("Family ID"), Gtk.CellRendererText(), text=4)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col3.set_resizable(True)
        col4.set_resizable(True)
        col5.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col3.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col4.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col5.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.treeview.append_column(col0)
        self.treeview.append_column(col1)
        self.treeview.append_column(col2)
        self.treeview.append_column(col3)
        self.treeview.append_column(col4)
        self.treeview.append_column(col5)
        self.treeselection = self.treeview.get_selection()
        self.treeview.connect("row-activated", self.rowactivated_cb)

        self.curr_fam = None
        people = self.db.get_person_handles()
        self.total = len(people)  # total number of people to process.
        self.count = 0  # current number of people completely processed
        self.loop = 0  # Number of loops found for GUI

        pset = OrderedDict()
        # pset is the handle list of persons from the current start of
        # exploration path to the current limit.  The use of OrderedDict
        # allows us to use it as a LIFO during recursion, as well as makes for
        # quick lookup.  If we find a loop, pset provides a nice way to get
        # the loop path.
        self.done = set()
        # self.done is the handle set of people that have been fully explored
        # and do NOT have loops in the decendent tree.  We use this to avoid
        # repeating work when we encounter one of these during the search.
        for person_handle in people:
            person = self.db.get_person_from_handle(person_handle)
            self.current = person
            self.parent = None
            self.descendants(person_handle, pset)

        # close the progress bar
        self.progress.close()

        self.show()

    def descendants(self, person_handle, pset):
        """
        Find the descendants of a given person.
        Returns False if a loop for the person is NOT found, True if loop found
        We use the return value to ensure a person is not put on done list if
        part of a loop
        """
        if person_handle in self.done:
            return False  # We have already verified no loops for this one
        if person_handle in pset:
            # We found one loop.
            # person_handle is child, self.parent, self.curr_fam valid
            # see if it has already been put into display
            person = self.db.get_person_from_handle(person_handle)
            pers_id = person.get_gramps_id()
            pers_name = _nd.display(person)
            parent_id = self.parent.get_gramps_id()
            parent_name = _nd.display(self.parent)
            value = (parent_id, parent_name, pers_id, pers_name, self.curr_fam)
            found = False
            for pth in range(len(self.model)):
                path = Gtk.TreePath(pth)
                treeiter = self.model.get_iter(path)
                find = (
                    self.model.get_value(treeiter, 0),
                    self.model.get_value(treeiter, 1),
                    self.model.get_value(treeiter, 2),
                    self.model.get_value(treeiter, 3),
                    self.model.get_value(treeiter, 4),
                )
                if find == value:
                    found = True  # This loop is in display model
                    break
            if not found:
                # Need to put loop in display model.
                self.loop += 1
                # place first node
                self.model.append(value + (str(self.loop),))
                state = 0
                # Now search for loop beginning.
                for hndl in pset.keys():
                    if hndl != person_handle and state == 0:
                        continue  # beginning not found
                    if state == 0:
                        state = 1  # found beginning, get first item to display
                        continue
                    # we have a good handle, now put item on display list
                    self.parent = person
                    person = self.db.get_person_from_handle(hndl)
                    # Get the family that is both parent/person
                    for fam_h in person.get_parent_family_handle_list():
                        if fam_h in self.parent.get_family_handle_list():
                            break
                    family = self.db.get_family_from_handle(fam_h)
                    fam_id = family.get_gramps_id()
                    pers_id = person.get_gramps_id()
                    pers_name = _nd.display(person)
                    parent_id = self.parent.get_gramps_id()
                    parent_name = _nd.display(self.parent)
                    value = (
                        parent_id,
                        parent_name,
                        pers_id,
                        pers_name,
                        fam_id,
                        str(self.loop),
                    )
                    self.model.append(value)
            return True
        # We are not part of loop (yet) so search descendents
        person = self.db.get_person_from_handle(person_handle)
        # put in the pset path list for recursive calls to find
        pset[person_handle] = None
        loop = False
        for family_handle in person.get_family_handle_list():
            family = self.db.get_family_from_handle(family_handle)
            if not family:
                # can happen with LivingProxyDb(PrivateProxyDb(db))
                continue
            for child_ref in family.get_child_ref_list():
                child_handle = child_ref.ref
                self.curr_fam = family.get_gramps_id()
                self.parent = person
                # if any descendants are part of loop, so is search person
                loop |= self.descendants(child_handle, pset)
        # we have completed search, we can pop the person off pset list
        person_handle, dummy = pset.popitem(last=True)

        if not loop:
            # person was not in loop, so add to done list and update progress
            self.done.add(person_handle)
            self.count += 1
            self.progress.set_header("%d/%d" % (self.count, self.total))
            self.progress.step()
            return False
        # person was in loop...
        return True

    def rowactivated_cb(self, treeview, path, column):
        """
        Called when a row is activated.
        """
        # first we need to check that the row corresponds to a person
        iter_ = self.model.get_iter(path)
        fam_id = self.model.get_value(iter_, 4)
        fam = self.dbstate.db.get_family_from_gramps_id(fam_id)
        if fam:
            try:
                EditFamily(self.dbstate, self.uistate, [], fam)
            except WindowActiveError:
                pass
            return True
        return False

    def on_help_clicked(self, obj):
        """
        Display the relevant portion of Gramps manual.
        """
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def close(self, *obj):
        ManagedWindow.close(self, *obj)


# ------------------------------------------------------------------------
#
# FindLoopOptions
#
# ------------------------------------------------------------------------
class FindLoopOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        """Initialize the options class"""
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.citationbase import CitationBase
from ...db import Database


# -------------------------------------------------------------------------
#
# HasSourceOf
#
# -------------------------------------------------------------------------
class HasSourceOfBase(Rule):
    """Rule that checks for objects that have a particular source."""

    labels = [_("Source ID:")]
    name = "Object with the <source>"
    category = _("Citation/source filters")
    description = "Matches objects who have a particular source"

    def prepare(self, db: Database, user):
        self.nosource = self.list[0] == ""
        if self.nosource:
            self.source_handle = None
        else:
            source = db.get_source_from_gramps_id(self.list[0])
            self.source_handle = source.handle if source else None

    def apply_to_one(self, db: Database, object: CitationBase) -> bool:
        if not self.source_handle:
            if self.nosource:
                # check whether the citation list is empty as a proxy for
                # there being no sources
                return len(object.get_all_citation_lists()) == 0
            else:
                return False
        else:
            for citation_handle in object.get_all_citation_lists():
                citation = db.get_citation_from_handle(citation_handle)
                if citation.source_handle == self.source_handle:
                    return True
            return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Martin Hawlisch, Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Dump Gender Statistics.

Tools/Debug/Dump Gender Statistics
"""

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gui.listmodel import ListModel, INTEGER
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.plug import tool

_GENDER = [_("female"), _("male"), _("unknown")]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class DumpGenderStats(tool.Tool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.label = _("Gender Statistics tool")
        tool.Tool.__init__(self, dbstate, options_class, name)

        stats_list = []
        for name, value in dbstate.db.genderStats.stats.items():
            stats_list.append(
                (name,) + value + (_GENDER[dbstate.db.genderStats.guess_gender(name)],)
            )

        if uistate:
            ManagedWindow.__init__(self, uistate, [], self.__class__)

            titles = [
                (_("Name"), 0, 100),
                (_("Male"), 1, 70, INTEGER),
                (_("Female"), 2, 70, INTEGER),
                (_("Unknown"), 3, 90, INTEGER),
                (_("Guess"), 4, 70),
            ]

            treeview = Gtk.TreeView()
            model = ListModel(treeview, titles)
            for entry in sorted(stats_list):
                model.add(entry, entry[0])

            s = Gtk.ScrolledWindow()
            s.add(treeview)
            dialog = Gtk.Dialog()
            dialog.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
            dialog.connect("response", self._response)
            dialog.vbox.pack_start(s, expand=True, fill=True, padding=0)
            self.set_window(dialog, None, self.label)
            self.setup_configs("interface.dumpgenderstats", 420, 300)
            self.show()

        else:
            if len(_("Name")) < 16:
                print(
                    "%s%s%s" % (_("Name"), " " * (16 - len(_("Name"))), _("Male")),
                    "\t%s" * 3 % (_("Female"), _("Unknown"), _("Guess")),
                )
            else:
                print(
                    _("Name"),
                    "\t%s" * 4 % (_("Male"), _("Female"), _("Unknown"), _("Guess")),
                )
            print()
            for entry in sorted(stats_list):
                if len(entry[0]) < 16:
                    print(
                        "%s%s%s" % (entry[0], " " * (16 - len(entry[0])), entry[1]),
                        "\t%s" * 3 % (entry[2:]),
                    )
                else:
                    print(entry[0], "\t%s" * 4 % (entry[1:]))

    def _response(self, obj, response_id):
        if response_id == Gtk.ResponseType.CLOSE:
            self.close()

    def build_menu_names(self, obj):
        return (self.label, None)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class DumpGenderStatsOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#

"""
Display all events on a particular day.
"""

from gramps.gen.simple import SimpleAccess, SimpleDoc, SimpleTable
from gramps.gui.plug.quick import QuickTable
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import Date


def get_ref(db, objclass, handle):
    """
    Looks up object in database
    """
    if objclass == "Person":
        ref = db.get_person_from_handle(handle)
    elif objclass == "Family":
        ref = db.get_family_from_handle(handle)
    elif objclass == "Event":
        ref = db.get_event_from_handle(handle)
    elif objclass == "Source":
        ref = db.get_source_from_handle(handle)
    elif objclass == "Place":
        ref = db.get_place_from_handle(handle)
    elif objclass == "Repository":
        ref = db.get_repository_from_handle(handle)
    else:
        ref = objclass
    return ref


def run(database, document, main_event):
    """
    Displays events on a specific date of an event (or date)

    Takes an Event or Date object
    """
    if isinstance(main_event, Date):
        main_date = main_event
    else:
        main_date = main_event.get_date_object()

    cal = main_date.get_calendar()

    # setup the simple access functions
    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)
    stab.set_link_col(3)
    yeartab = QuickTable(sdb)
    yeartab.set_link_col(3)
    histab = QuickTable(sdb)
    histab.set_link_col(3)

    # display the title
    sdoc.title(_("Events of %(date)s") % {"date": sdb.date_string(main_date)})
    sdoc.paragraph("")
    stab.columns(_("Date"), _("Type"), _("Place"), _("Reference"))
    yeartab.columns(_("Date"), _("Type"), _("Place"), _("Reference"))
    histab.columns(_("Date"), _("Type"), _("Place"), _("Reference"))

    for event in database.iter_events():
        date = event.get_date_object()
        date.convert_calendar(cal)
        if date.get_year() == 0:
            continue
        if (
            date.get_year() == main_date.get_year()
            and date.get_month() == main_date.get_month()
            and date.get_day() == main_date.get_day()
        ):
            for objclass, handle in database.find_backlink_handles(event.handle):
                ref = get_ref(database, objclass, handle)
                stab.row(date, sdb.event_type(event), sdb.event_place(event), ref)
        elif (
            date.get_month() == main_date.get_month()
            and date.get_day() == main_date.get_day()
            and date.get_month() != 0
        ):
            for objclass, handle in database.find_backlink_handles(event.handle):
                ref = get_ref(database, objclass, handle)
                histab.row(date, sdb.event_type(event), sdb.event_place(event), ref)
        elif date.get_year() == main_date.get_year():
            for objclass, handle in database.find_backlink_handles(event.handle):
                ref = get_ref(database, objclass, handle)
                yeartab.row(date, sdb.event_type(event), sdb.event_place(event), ref)

    document.has_data = False
    if stab.get_row_count() > 0:
        document.has_data = True
        sdoc.paragraph(_("Events on this exact date"))
        stab.write(sdoc)
    else:
        sdoc.paragraph(_("No events on this exact date"))
        sdoc.paragraph("")
    sdoc.paragraph("")

    if histab.get_row_count() > 0:
        document.has_data = True
        sdoc.paragraph(_("Other events on this month/day in history"))
        histab.write(sdoc)
    else:
        sdoc.paragraph(_("No other events on this month/day in history"))
        sdoc.paragraph("")
    sdoc.paragraph("")

    if yeartab.get_row_count() > 0:
        document.has_data = True
        sdoc.paragraph(_("Other events in %(year)d") % {"year": main_date.get_year()})
        yeartab.write(sdoc)
    else:
        sdoc.paragraph(
            _("No other events in %(year)d") % {"year": main_date.get_year()}
        )
        sdoc.paragraph("")
    sdoc.paragraph("")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# -------------------------------------------------------------------------
#
# Python
#
# -------------------------------------------------------------------------
from html import escape


# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gramps.gen.config import config
from gramps.gen.utils.string import conf_strings
from gramps.gen.datehandler import get_date, get_date_valid
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Globals
#
# -------------------------------------------------------------------------
invalid_date_format = config.get("preferences.invalid-date-format")


# -------------------------------------------------------------------------
#
# CitationModel
#
# -------------------------------------------------------------------------
class CitationRefModel(Gtk.ListStore):
    def __init__(self, citation_list, db):
        Gtk.ListStore.__init__(
            self, str, str, str, str, str, str, str, bool, str, int, str
        )
        self.db = db
        dbgsfh = self.db.get_source_from_handle
        for handle in citation_list:
            citation = self.db.get_citation_from_handle(handle)
            src = dbgsfh(citation.get_reference_handle())
            confidence = citation.get_confidence_level()
            self.append(
                row=[
                    src.title,
                    src.author,
                    self.column_date(citation),
                    src.get_publication_info(),
                    _(conf_strings[confidence]),
                    citation.page,
                    citation.gramps_id,
                    citation.get_privacy(),
                    self.column_sort_date(citation),
                    self.column_sort_confidence(citation),
                    handle,
                ]
            )

    def column_date(self, citation):
        retval = get_date(citation)
        if not get_date_valid(citation):
            return invalid_date_format % escape(retval)
        else:
            return retval

    def column_sort_date(self, citation):
        date = citation.get_date_object()
        if date:
            return "%09d" % date.get_sort_value()
        else:
            return ""

    def column_sort_confidence(self, citation):
        confidence = citation.get_confidence_level()
        return confidence



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
# Copyright (C) 2009 Douglas S. Blank
# Copyright (C) 2012 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

## Based on the paper:
##   http://www.cs.utah.edu/~draperg/research/fanchart/draperg_FHT08.pdf
## and the applet:
##   http://www.cs.utah.edu/~draperg/research/fanchart/demo/

## Found by redwood:
## http://www.gramps-project.org/bugs/view.php?id=2611

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import math
import cairo

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug.report.utils import find_spouse
from ..utils import hex_to_rgb
from .fanchart import (
    FanChartBaseWidget,
    FanChartGrampsGUI,
    PAD_PX,
    TRANSLATE_PX,
    FORM_CIRCLE,
    FORM_HALFCIRCLE,
    FORM_QUADRANT,
    NORMAL,
    EXPANDED,
    COLLAPSED,
    BACKGROUND_GRAD_GEN,
    BACKGROUND_GRAD_AGE,
    BACKGROUND_GRAD_PERIOD,
    CHILDRING_WIDTH,
)

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

PIXELS_PER_GENPERSON_RATIO = 0.55  # ratio of generation radius for person
# (rest for partner)
PIXELS_PER_GEN_SMALL = 80
PIXELS_PER_GEN_LARGE = 160
N_GEN_SMALL = 4
PIXELS_PER_GENFAMILY = 25  # size of radius for family
PIXELS_PER_RECLAIM = 4  # size of the radius of pixels taken from family
# to reclaim space
PIXELS_PARTNER_GAP = 0  # Padding between someone and his partner
PIXELS_CHILDREN_GAP = 5  # Padding between generations
PARENTRING_WIDTH = 12  # width of the parent ring inside the person

ANGLE_CHEQUI = 0  # Algorithm with homogeneous children distribution
ANGLE_WEIGHT = 1  # Algorithm for angle computation based on nr of descendants

TYPE_BOX_NORMAL = 0
TYPE_BOX_FAMILY = 1

# -------------------------------------------------------------------------
#
# FanChartDescWidget
#
# -------------------------------------------------------------------------


class FanChartDescWidget(FanChartBaseWidget):
    """
    Interactive Fan Chart Widget.
    """

    CENTER = 60  # we require a larger center as CENTER includes the 1st partner

    def __init__(self, dbstate, uistate, callback_popup=None):
        """
        Fan Chart Widget. Handles visualization of data in self.data.
        See main() of FanChartGramplet for example of model format.
        """
        self.gen2people = {}
        self.gen2fam = {}
        self.rootangle_rad = []
        self.handle2desc = {}
        self.famhandle2desc = {}
        self.handle2fam = {}
        self.innerring = []
        self.angle = {}
        self.set_values(
            None,
            9,
            True,
            True,
            BACKGROUND_GRAD_GEN,
            "Sans",
            "#0000FF",
            "#FF0000",
            None,
            0.5,
            FORM_CIRCLE,
            ANGLE_WEIGHT,
            "#888a85",
            False,
        )
        FanChartBaseWidget.__init__(self, dbstate, uistate, callback_popup)

    def set_values(
        self,
        root_person_handle,
        maxgen,
        flipupsidedownname,
        twolinename,
        background,
        fontdescr,
        grad_start,
        grad_end,
        filtr,
        alpha_filter,
        form,
        angle_algo,
        dupcolor,
        showid,
    ):
        """
        Reset the values to be used:

        :param root_person_handle: person to show
        :param maxgen: maximum generations to show
        :param flipupsidedownname: flip name on the left of the fanchart
                                   for the display of person's name
        :param background: config setting of which background procedure to use
        :type background: int
        :param fontdescr: string describing the font to use
        :param grad_start: colors to use for background procedure
        :param grad_end: colors to use for background procedure
        :param filtr: the person filter to apply to the people in the chart
        :param alpha_filter: the alpha transparency value (0-1) to apply to
                             filtered out data
        :param form: the ``FORM_`` constant for the fanchart
        :param angle_algo: alorithm to use to calculate the sizes of the boxes
        :param dupcolor: color to use for people or families that occur a second
                         or more time
        :param showid: to show the gramps_id or not
        """
        self.rootpersonh = root_person_handle
        self.generations = maxgen
        self.background = background
        self.fontdescr = fontdescr
        self.grad_start = grad_start
        self.grad_end = grad_end
        self.filter = filtr
        self.alpha_filter = alpha_filter
        self.form = form
        self.anglealgo = angle_algo
        self.dupcolor = hex_to_rgb(dupcolor)
        self.childring = False
        self.flipupsidedownname = flipupsidedownname
        self.twolinename = twolinename
        self.showid = showid

    def set_generations(self):
        """
        Set the generations to max, and fill data structures with initial data.
        """

        if self.form == FORM_CIRCLE:
            self.rootangle_rad = [math.radians(0), math.radians(360)]
        elif self.form == FORM_HALFCIRCLE:
            self.rootangle_rad = [math.radians(90), math.radians(90 + 180)]
        elif self.form == FORM_QUADRANT:
            self.rootangle_rad = [math.radians(180), math.radians(270)]

        self.handle2desc = {}
        self.famhandle2desc = {}
        self.handle2fam = {}
        self.gen2people = {}
        self.gen2fam = {}
        self.innerring = []
        self.gen2people[0] = [
            (None, False, 0, 2 * math.pi, 0, 0, [], NORMAL)
        ]  # no center person
        self.gen2fam[0] = []  # no families
        self.angle = {}
        self.angle[-2] = []
        for i in range(1, self.generations + 1):
            self.gen2fam[i] = []
            self.gen2people[i] = []
        self.gen2people[self.generations] = []  # indication of more children

    def _fill_data_structures(self):
        self.set_generations()
        if not self.rootpersonh:
            return
        person = self.dbstate.db.get_person_from_handle(self.rootpersonh)
        if not person:
            # nothing to do, just return
            return

        # person, duplicate or not, start angle, slice size,
        #                   text, parent pos in fam, nrfam, userdata, status
        self.gen2people[0] = [[person, False, 0, 2 * math.pi, 0, 0, [], NORMAL]]
        self.handle2desc[self.rootpersonh] = 0
        # fill in data for the parents
        self.innerring = []
        handleparents = []
        family_handle_list = person.get_parent_family_handle_list()
        if family_handle_list:
            for family_handle in family_handle_list:
                family = self.dbstate.db.get_family_from_handle(family_handle)
                if not family:
                    continue
                for hparent in [family.get_father_handle(), family.get_mother_handle()]:
                    if hparent and hparent not in handleparents:
                        parent = self.dbstate.db.get_person_from_handle(hparent)
                        if parent:
                            self.innerring.append((parent, []))
                            handleparents.append(hparent)

        # recursively fill in the datastructures:
        nrdesc = self._rec_fill_data(0, person, 0, self.generations)
        self.handle2desc[person.handle] += nrdesc
        self._compute_angles(*self.rootangle_rad)

    def _rec_fill_data(self, gen, person, pos, maxgen):
        """
        Recursively fill in the data
        """
        totdesc = 0
        marriage_handle_list = person.get_family_handle_list()
        self.gen2people[gen][pos][5] = len(marriage_handle_list)
        for family_handle in marriage_handle_list:
            totdescfam = 0
            family = self.dbstate.db.get_family_from_handle(family_handle)

            spouse_handle = find_spouse(person, family)
            if spouse_handle:
                spouse = self.dbstate.db.get_person_from_handle(spouse_handle)
            else:
                spouse = None
            # family may occur via father and via mother in the chart, only
            # first to show and count.
            fam_duplicate = family_handle in self.famhandle2desc
            # family, duplicate or not, start angle, slice size,
            #   spouse pos in gen, nrchildren, userdata, parnter, status
            self.gen2fam[gen].append(
                [family, fam_duplicate, 0, 0, pos, 0, [], spouse, NORMAL]
            )
            posfam = len(self.gen2fam[gen]) - 1

            if not fam_duplicate and gen < maxgen - 1:
                nrchild = len(family.get_child_ref_list())
                self.gen2fam[gen][posfam][5] = nrchild
                for child_ref in family.get_child_ref_list():
                    chld = self.dbstate.db.get_person_from_handle(child_ref.ref)
                    child_dup = child_ref.ref in self.handle2desc
                    if not child_dup:
                        # mark this child as processed
                        self.handle2desc[child_ref.ref] = 0
                    # person, duplicate or not, start angle, slice size,
                    #         parent pos in fam, nrfam, userdata, status
                    self.gen2people[gen + 1].append(
                        [chld, child_dup, 0, 0, posfam, 0, [], NORMAL]
                    )
                    totdescfam += 1  # add this person as descendant
                    pospers = len(self.gen2people[gen + 1]) - 1
                    if not child_dup:
                        nrdesc = self._rec_fill_data(gen + 1, chld, pospers, maxgen)
                        self.handle2desc[child_ref.ref] += nrdesc
                        # add children of him as descendants
                        totdescfam += nrdesc
            if not fam_duplicate:
                self.famhandle2desc[family_handle] = totdescfam
            totdesc += totdescfam
        return totdesc

    def _compute_angles(self, start_rad, stop_rad):
        """
        Compute the angles of the boxes
        """
        # first we compute the size of the slice.
        # set angles root person
        start, portion = start_rad, stop_rad - start_rad
        nr_gen = len(self.gen2people) - 1
        # Fill in central person angles
        gen = 0
        data = self.gen2people[gen][0]
        data[2] = start
        data[3] = portion
        for gen in range(0, nr_gen):
            prevpartnerdatahandle = None
            offset = 0
            for data_fam in self.gen2fam[gen]:  # for each partner/fam of gen-1
                # obtain start and stop from the people of this partner
                persondata = self.gen2people[gen][data_fam[4]]
                dupfam = data_fam[1]
                if dupfam:
                    # we don't show again the descendants here
                    nrdescfam = 0
                else:
                    nrdescfam = self.famhandle2desc[data_fam[0].handle]
                nrdescperson = self.handle2desc[persondata[0].handle]
                nrfam = persondata[5]
                personstart, personslice = persondata[2:4]
                if prevpartnerdatahandle != persondata[0].handle:
                    # partner of a new person: reset the offset
                    offset = 0
                    prevpartnerdatahandle = persondata[0].handle
                portion = personslice / (nrdescperson + nrfam) * (nrdescfam + 1)
                if data_fam[8] == COLLAPSED:
                    portion = 0
                elif data_fam[8] == EXPANDED:
                    portion = personslice

                data_fam[2] = personstart + offset
                data_fam[3] = portion
                offset += portion

            ##                if nrdescperson == 0:
            ##                    #no offspring, draw as large as fraction of
            ##                    #nr families
            ##                    nrfam = persondata[6]
            ##                    portion = personslice/nrfam
            ##                    data_fam[2] = personstart + offset
            ##                    data_fam[3] = portion
            ##                    offset += portion
            ##                elif nrdescfam == 0:
            ##                    #no offspring this family, but there is another
            ##                    #family. We draw this as a weight of 1
            ##                    nrfam = persondata[6]
            ##                    portion = (personslice/(nrdescperson + nrfam - 1) *
            ##                               (nrdescfam + 1))
            ##                    data_fam[2] = personstart + offset
            ##                    data_fam[3] = portion
            ##                    offset += portion
            ##                else:
            ##                    #this family has offspring. We give it space for it's
            ##                    #weight in offspring
            ##                    nrfam = persondata[6]
            ##                    portion = (personslice/(nrdescperson + nrfam - 1) *
            ##                               (nrdescfam + 1))
            ##                    data_fam[2] = personstart + offset
            ##                    data_fam[3] = portion
            ##                    offset += portion

            prevfamdatahandle = None
            offset = 0
            for persondata in self.gen2people[gen + 1] if gen < nr_gen else []:
                # obtain start and stop of family this is child of
                parentfamdata = self.gen2fam[gen][persondata[4]]
                nrdescfam = 0
                if not parentfamdata[1]:
                    nrdescfam = self.famhandle2desc[parentfamdata[0].handle]
                nrdesc = 0
                if not persondata[1]:
                    nrdesc = self.handle2desc[persondata[0].handle]
                famstart = parentfamdata[2]
                famslice = parentfamdata[3]
                nrchild = parentfamdata[5]
                # now we divide this slice to the weight of children,
                # adding one for every child
                if self.anglealgo == ANGLE_CHEQUI:
                    portion = famslice / nrchild
                elif self.anglealgo == ANGLE_WEIGHT:
                    portion = famslice / (nrdescfam) * (nrdesc + 1)
                else:
                    raise NotImplementedError(
                        "Unknown angle algorithm %d" % self.anglealgo
                    )
                if prevfamdatahandle != parentfamdata[0].handle:
                    # reset the offset
                    offset = 0
                    prevfamdatahandle = parentfamdata[0].handle
                if persondata[7] == COLLAPSED:
                    portion = 0
                elif persondata[7] == EXPANDED:
                    portion = famslice
                persondata[2] = famstart + offset
                persondata[3] = portion
                offset += portion

    def nrgen(self):
        """
        compute the number of generations present
        """
        for gen in range(self.generations - 1, 0, -1):
            if self.gen2people[gen]:
                return gen + 1
        return 1

    def halfdist(self):
        """
        Compute the half radius of the circle
        """
        return self.get_radiusinout_for_generation(self.nrgen())[1]

    def get_radiusinout_for_generation(self, generation):
        """
        Get the in and out radius for the generation
        """
        radius_first_gen = 14  # fudged to make inner circle a bit tighter
        if generation < N_GEN_SMALL:
            radius_start = PIXELS_PER_GEN_SMALL * generation + radius_first_gen
            return (radius_start, radius_start + PIXELS_PER_GEN_SMALL)
        else:
            radius_start = (
                PIXELS_PER_GEN_SMALL * N_GEN_SMALL
                + PIXELS_PER_GEN_LARGE * (generation - N_GEN_SMALL)
                + radius_first_gen
            )
            return (radius_start, radius_start + PIXELS_PER_GEN_LARGE)

    def get_radiusinout_for_gen_pair(self, generation):
        """
        Get the in and out radius for the father and mother
        """
        radiusin, radiusout = self.get_radiusinout_for_generation(generation)
        radius_spread = radiusout - radiusin - PIXELS_CHILDREN_GAP - PIXELS_PARTNER_GAP

        radiusin_pers = radiusin + PIXELS_CHILDREN_GAP
        radiusout_pers = radiusin_pers + PIXELS_PER_GENPERSON_RATIO * radius_spread
        radiusin_partner = radiusout_pers + PIXELS_PARTNER_GAP
        radiusout_partner = radiusout
        return (radiusin_pers, radiusout_pers, radiusin_partner, radiusout_partner)

    def people_generator(self):
        """
        a generator over all people outside of the core person
        """
        for generation in range(self.generations):
            for data in self.gen2people[generation]:
                yield (data[0], data[6])
        for generation in range(self.generations):
            for data in self.gen2fam[generation]:
                yield (data[7], data[6])

    def innerpeople_generator(self):
        """
        a generator over all people inside of the core person
        """
        for parentdata in self.innerring:
            parent, userdata = parentdata
            yield (parent, userdata)

    def draw(self, ctx=None, scale=1.0):
        """
        The main method to do the drawing.
        If ctx is given, we assume we draw draw raw on the cairo context ctx
        To draw in GTK3 and use the allocation, set ctx=None.
        Note: when drawing for display, to counter a Gtk issue with scrolling
        or resizing the drawing window, we draw on a surface, then copy to the
        drawing context when the Gtk 'draw' signal arrives.
        """
        # first do size request of what we will need
        halfdist = self.halfdist()
        if not ctx:  # Display
            if self.form == FORM_CIRCLE:
                size_w = size_h = 2 * halfdist
            elif self.form == FORM_HALFCIRCLE:
                size_w = 2 * halfdist
                size_h = halfdist + self.CENTER + PAD_PX
            elif self.form == FORM_QUADRANT:
                size_w = size_h = halfdist + self.CENTER + PAD_PX

            size_w_a = self.get_allocated_width()
            size_h_a = self.get_allocated_height()
            self.set_size_request(max(size_w, size_w_a), max(size_h, size_h_a))
            size_w = self.get_allocated_width()
            size_h = self.get_allocated_height()
            scale_factor = self.get_scale_factor()
            self.surface = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, size_w * scale_factor, size_h * scale_factor
            )
            self.surface.set_device_scale(scale_factor, scale_factor)
            ctx = cairo.Context(self.surface)
            self.center_xy = self.center_xy_from_delta()
            ctx.translate(*self.center_xy)
        else:  # printing
            if self.form == FORM_QUADRANT:
                self.center_xy = self.CENTER, halfdist
            else:
                self.center_xy = halfdist, halfdist
            ctx.scale(scale, scale)
            ctx.translate(*self.center_xy)

        ctx.save()
        # Draw center person:
        (
            person,
            dup,
            start,
            portion,
            dummy_parentfampos,
            dummy_nrfam,
            userdata,
            status,
        ) = self.gen2people[0][0]
        if person:
            # r, g, b, a = self.background_box(person, 0, userdata)
            (
                radiusin_pers,
                radiusout_pers,
                radiusin_partner,
                radiusout_partner,
            ) = self.get_radiusinout_for_gen_pair(0)
            if not self.innerring:
                radiusin_pers = TRANSLATE_PX
            self.draw_person(
                ctx,
                person,
                radiusin_pers,
                radiusout_pers,
                math.pi / 2,
                math.pi / 2 + 2 * math.pi,
                0,
                False,
                userdata,
                is_central_person=True,
            )
            # draw center to move chart
            ctx.set_source_rgb(0, 0, 0)  # black
            ctx.move_to(TRANSLATE_PX, 0)
            ctx.arc(0, 0, TRANSLATE_PX, 0, 2 * math.pi)
            if self.innerring:  # has at least one parent
                ctx.fill()
                self.draw_innerring_people(ctx)
            else:
                ctx.stroke()
        # now write all the families and children
        ctx.rotate(self.rotate_value * math.pi / 180)
        for gen in range(self.generations):
            (
                radiusin_pers,
                radiusout_pers,
                radiusin_partner,
                radiusout_partner,
            ) = self.get_radiusinout_for_gen_pair(gen)
            if gen > 0:
                for pdata in self.gen2people[gen]:
                    # person, duplicate or not, start angle, slice size,
                    #             parent pos in fam, nrfam, userdata, status
                    (
                        pers,
                        dup,
                        start,
                        portion,
                        dummy_pospar,
                        dummy_nrfam,
                        userdata,
                        status,
                    ) = pdata
                    if status != COLLAPSED:
                        self.draw_person(
                            ctx,
                            pers,
                            radiusin_pers,
                            radiusout_pers,
                            start,
                            start + portion,
                            gen,
                            dup,
                            userdata,
                            thick=status != NORMAL,
                        )
            # if gen < self.generations-1:
            for famdata in self.gen2fam[gen]:
                # family, duplicate or not, start angle, slice size,
                #       spouse pos in gen, nrchildren, userdata, status
                (
                    fam,
                    dup,
                    start,
                    portion,
                    dummy_posfam,
                    dummy_nrchild,
                    userdata,
                    partner,
                    status,
                ) = famdata
                if status != COLLAPSED:
                    more_pers_flag = (
                        gen == self.generations - 1 and fam.get_child_ref_list()
                    )
                    self.draw_person(
                        ctx,
                        partner,
                        radiusin_partner,
                        radiusout_partner,
                        start,
                        start + portion,
                        gen,
                        dup,
                        userdata,
                        thick=(status != NORMAL),
                        has_moregen_indicator=more_pers_flag,
                    )
        ctx.restore()

        if self.background in [BACKGROUND_GRAD_AGE, BACKGROUND_GRAD_PERIOD]:
            self.draw_gradient_legend(ctx, halfdist)

    def cell_address_under_cursor(self, curx, cury):
        """
        Determine the cell address in the fan under the cursor
        position x and y.
        None if outside of diagram
        """
        radius, rads, raw_rads = self.cursor_to_polar(curx, cury, get_raw_rads=True)

        btype = TYPE_BOX_NORMAL
        if radius < TRANSLATE_PX:
            return None
        elif (
            self.innerring
            and self.angle[-2]
            and radius < CHILDRING_WIDTH + TRANSLATE_PX
        ):
            generation = -2  # indication of one of the children
        elif radius < self.CENTER:
            generation = 0
        else:
            generation = None
            for gen in range(self.generations):
                (
                    radiusin_pers,
                    radiusout_pers,
                    radiusin_partner,
                    radiusout_partner,
                ) = self.get_radiusinout_for_gen_pair(gen)
                if radiusin_pers <= radius <= radiusout_pers:
                    generation, btype = gen, TYPE_BOX_NORMAL
                    break
                if radiusin_partner <= radius <= radiusout_partner:
                    generation, btype = gen, TYPE_BOX_FAMILY
                    break

        # find what person is in this position:
        selected = None
        if not (generation is None) and generation > 0:
            selected = self.personpos_at_angle_btype(generation, rads, btype)
        elif generation == -2:
            for idx, (start, stop, dummy) in enumerate(self.angle[generation]):
                if self.radian_in_bounds(start, raw_rads, stop):
                    selected = idx
                    break
        if generation is None or selected is None:
            return None
        return generation, selected, btype

    def draw_innerring_people(self, ctx):
        """
        Draw the innerring person
        """
        ctx.move_to(TRANSLATE_PX + CHILDRING_WIDTH, 0)
        ctx.set_source_rgb(0, 0, 0)  # black
        ctx.set_line_width(1)
        ctx.arc(0, 0, TRANSLATE_PX + CHILDRING_WIDTH, 0, 2 * math.pi)
        ctx.stroke()
        nrparent = len(self.innerring)
        # Y axis is downward. positve angles are hence clockwise
        startangle = math.pi
        if nrparent <= 2:
            angleinc = math.pi
        elif nrparent <= 4:
            angleinc = math.pi / 2
        else:
            angleinc = 2 * math.pi / nrparent
        for data in self.innerring:
            self.draw_innerring(ctx, data[0], data[1], startangle, angleinc)
            startangle += angleinc

    def personpos_at_angle_btype(self, generation, rads, btype):
        """
        returns the person in generation generation at angle.
        """
        selected = None
        datas = None
        if btype == TYPE_BOX_NORMAL:
            if generation == 0:
                return 0  # central person is always ok !
            datas = self.gen2people[generation]
        elif btype == TYPE_BOX_FAMILY:
            datas = self.gen2fam[generation]
        else:
            return None
        for idx, pdata in enumerate(datas):
            # person, duplicate or not, start angle, slice size,
            #             parent pos in fam, nrfam, userdata, status
            start, stop = pdata[2], pdata[2] + pdata[3]
            if self.radian_in_bounds(start, rads, stop):
                selected = idx
                break
        return selected

    def person_at(self, cell_address):
        """
        returns the person at generation, pos, btype
        """
        generation, pos, btype = cell_address
        if generation == -2:
            person, dummy_userdata = self.innerring[pos]
        elif btype == TYPE_BOX_NORMAL:
            # person, duplicate or not, start angle, slice size,
            #                   parent pos in fam, nrfam, userdata, status
            person = self.gen2people[generation][pos][0]
        elif btype == TYPE_BOX_FAMILY:
            # family, duplicate or not, start angle, slice size,
            #       spouse pos in gen, nrchildren, userdata, person, status
            person = self.gen2fam[generation][pos][7]
        return person

    def family_at(self, cell_address):
        """
        returns the family at generation, pos, btype
        """
        generation, pos, btype = cell_address
        if pos is None or btype == TYPE_BOX_NORMAL or generation < 0:
            return None
        return self.gen2fam[generation][pos][0]

    def do_mouse_click(self):
        """
        no drag occured, expand or collapse the section
        """
        self.toggle_cell_state(self._mouse_click_cell_address)
        self._compute_angles(*self.rootangle_rad)
        self._mouse_click = False
        self.draw()
        self.queue_draw()

    def toggle_cell_state(self, cell_address):
        """
        Toggle the person cell (EXPAND/COLLAPSE)
        """
        generation, selected, btype = cell_address
        if generation < 1:
            return
        if btype == TYPE_BOX_NORMAL:
            data = self.gen2people[generation][selected]
            parpos = data[4]
            status = data[7]
            if status == NORMAL:
                # should be expanded, rest collapsed
                for entry in self.gen2people[generation]:
                    if entry[4] == parpos:
                        entry[7] = COLLAPSED
                data[7] = EXPANDED
            else:
                # is expanded, set back to normal
                for entry in self.gen2people[generation]:
                    if entry[4] == parpos:
                        entry[7] = NORMAL
        if btype == TYPE_BOX_FAMILY:
            data = self.gen2fam[generation][selected]
            parpos = data[4]
            status = data[8]
            if status == NORMAL:
                # should be expanded, rest collapsed
                for entry in self.gen2fam[generation]:
                    if entry[4] == parpos:
                        entry[8] = COLLAPSED
                data[8] = EXPANDED
            else:
                # is expanded, set back to normal
                for entry in self.gen2fam[generation]:
                    if entry[4] == parpos:
                        entry[8] = NORMAL


class FanChartDescGrampsGUI(FanChartGrampsGUI):
    """class for functions fanchart GUI elements will need in Gramps"""

    def main(self):
        """
        Fill the data structures with the active data. This initializes all
        data.
        """
        root_person_handle = self.get_active("Person")
        self.fan.set_values(
            root_person_handle,
            self.maxgen,
            self.flipupsidedownname,
            self.twolinename,
            self.background,
            self.fonttype,
            self.grad_start,
            self.grad_end,
            self.generic_filter,
            self.alpha_filter,
            self.form,
            self.angle_algo,
            self.dupcolor,
            self.showid,
        )
        self.fan.reset()
        self.fan.draw()
        self.fan.queue_draw()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017  Paul Culley <paulr2787_at_gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for full or partial name matches based on soundex.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.nameorigintype import NameOriginType
from ....soundex import soundex
from .. import Rule

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasSoundexName
#
# -------------------------------------------------------------------------
class HasSoundexName(Rule):
    """
    Rule that checks for full or partial name matches based on soundex.
    """

    labels = [_("Name:")]
    name = _("Soundex match of People with the <name>")
    description = _(
        "Soundex Match of people with a specified name. First name, Surname, "
        "Call name, and Nickname are searched in primary and alternate names."
    )
    category = _("General filters")
    allow_regex = False

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.soundex = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things we only want to do once.
        """
        if self.list[0]:
            self.soundex = soundex(self.list[0])

    def apply_to_one(self, _db: Database, obj: Person) -> bool:
        """
        Apply the rule. Return True on a match.
        """
        for name in [obj.primary_name] + obj.alternate_names:
            if self._match_name(name):
                return True
        return False

    def _match_name(self, name):
        """
        Match a name against the soundex.
        """
        if self.soundex:
            if soundex(str(name.first_name)) == self.soundex:
                return True
            if soundex(str(name.get_surname())) == self.soundex:
                return True
            if soundex(str(name.call)) == self.soundex:
                return True
            if soundex(str(name.nick)) == self.soundex:
                return True
            if soundex(str(name.famnick)) == self.soundex:
                return True
            for surname in name.surname_list:
                if self._match_surname(surname):
                    return True
        return False

    def _match_surname(self, surname):
        """
        Match a surname against the soundex.
        """
        if soundex(str(surname.get_surname())) == self.soundex:
            return True
        if int(surname.origintype.value) == NameOriginType.PATRONYMIC:
            if soundex(str(surname.surname)) == self.soundex:
                return True
        return False



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010,2015  Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2018       Paul D.Smith
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import logging

from gramps.gen.const import GRAMPS_LOCALE as glocale

LOG = logging.getLogger(".NarrativeWeb.BuchheimTree")

_ = glocale.translation.sgettext


# ------------------------------------------------------------
#
# DrawTree - a Buchheim draw tree which implements the
#   tree drawing algorithm of:
#
#   Improving Walker's algorithm to Run in Linear Time
#   Christoph Buchheim, Michael Juenger, and Sebastian Leipert
#
#   Also see:
#
#   Positioning Nodes for General Trees
#   John Q. Walker II
#
# The following modifications are noted:
#
# - The root node is 'west' according to the later nomenclature
#   employed by Walker with the nodes stretching 'east'
# - This reverses the X & Y co-originates of the Buchheim paper
# - The algorithm has been tweaked to track the maximum X and Y
#   as 'width' and 'height' to aid later layout
# - The Buchheim examples track a string identifying the actual
#   node but this implementation tracks the handle of the
#   DB node identifying the person in the Gramps DB.  This is done
#   to minimize occupancy at any one time.
# ------------------------------------------------------------
class DrawTree(object):
    def __init__(self, tree, parent=None, depth=0, number=1):
        self.coord_x = -1.0
        self.coord_y = depth
        self.width = self.coord_x
        self.height = self.coord_y
        self.tree = tree
        self.children = [
            DrawTree(c, self, depth + 1, i + 1) for i, c in enumerate(tree.children)
        ]
        self.parent = parent
        self.thread = None
        self.mod = 0
        self.ancestor = self
        self.change = self.shift = 0
        self._lmost_sibling = None
        # this is the number of the node in its group of siblings 1..n
        self.number = number

    def left(self):
        """
        Return the left most child if it exists.
        """
        return self.thread or len(self.children) and self.children[0]

    def right(self):
        """
        Return the rightmost child if it exists.
        """
        return self.thread or len(self.children) and self.children[-1]

    def lbrother(self):
        """
        Return the sibling to the left of this one.
        """
        brother = None
        if self.parent:
            for node in self.parent.children:
                if node == self:
                    return brother
                else:
                    brother = node
        return brother

    def get_lmost_sibling(self):
        """
        Return the leftmost sibling.
        """
        if not self._lmost_sibling and self.parent and self != self.parent.children[0]:
            self._lmost_sibling = self.parent.children[0]
        return self._lmost_sibling

    lmost_sibling = property(get_lmost_sibling)

    def __str__(self):
        return "%s: x=%s mod=%s" % (self.tree, self.coord_x, self.mod)

    def __repr__(self):
        return self.__str__()

    def handle(self):
        """
        Return the handle of the tree, which is whatever we stored as
        in the tree to reference out data.
        """
        return self.tree.handle


def buchheim(tree, node_width, h_separation, node_height, v_separation):
    """
    Calculate the position of elements of the graph given a minimum
    generation width separation and minimum generation height separation.
    """
    draw_tree = firstwalk(DrawTree(tree), node_height, v_separation)
    min_x = second_walk(draw_tree, 0, node_width + h_separation, 0)
    if min_x < 0:
        third_walk(draw_tree, 0 - min_x)
    top = get_min_coord_y(draw_tree)
    height = get_max_coord_y(draw_tree)

    return (draw_tree, top, height)


def get_min_coord_y(tree, min_value=100.0):
    """Get the minimum coord_y"""
    if tree.coord_y < min_value:
        min_value = tree.coord_y
    for child in tree.children:
        min_v = get_min_coord_y(child, min_value)
        if min_value > min_v:
            min_value = min_v
    return min_value


def get_max_coord_y(tree, max_value=0.0):
    """Get the maximum coord_y"""
    if tree.coord_y > max_value:
        max_value = tree.coord_y
    for child in tree.children:
        max_v = get_max_coord_y(child, max_value)
        if max_value < max_v:
            max_value = max_v
    return max_value


def third_walk(tree, adjust):
    """
    The tree has have wandered into 'negative' co-ordinates so bring it back
    into the piositive domain.
    """
    tree.coord_x += adjust
    tree.width = max(tree.width, tree.coord_x)
    for child in tree.children:
        third_walk(child, adjust)


def firstwalk(tree, node_height, v_separation):
    """
    Determine horizontal positions.
    """
    if not tree.children:
        if tree.lmost_sibling:
            tree.coord_y = tree.lbrother().coord_y + node_height + v_separation
        else:
            tree.coord_y = 0.0
    else:
        default_ancestor = tree.children[0]
        for child in tree.children:
            firstwalk(child, node_height, v_separation)
            default_ancestor = apportion(
                child, default_ancestor, node_height + v_separation
            )
            tree.height = max(tree.height, child.height)
            assert tree.width >= child.width
        execute_shifts(tree)

        midpoint = (tree.children[0].coord_y + tree.children[-1].coord_y) / 2

        brother = tree.lbrother()
        if brother:
            tree.coord_y = brother.coord_y + node_height + v_separation
            tree.mod = tree.coord_y - midpoint
        else:
            tree.coord_y = midpoint

    assert tree.width >= tree.coord_x
    tree.height = max(tree.height, tree.coord_y)
    return tree


def apportion(tree, default_ancestor, v_separation):
    """
    Figure out relative positions of node in a tree.
    """
    brother = tree.lbrother()
    if brother is not None:
        # in buchheim notation:
        # i == inner; o == outer; r == right; l == left; r = +; l = -
        vir = vor = tree
        vil = brother
        vol = tree.lmost_sibling
        sir = sor = tree.mod
        sil = vil.mod
        sol = vol.mod
        while vil.right() and vir.left():
            vil = vil.right()
            vir = vir.left()
            vol = vol.left()
            vor = vor.right()
            vor.ancestor = tree
            shift = (vil.coord_y + sil) - (vir.coord_y + sir) + v_separation
            if shift > 0:
                move_subtree(ancestor(vil, tree, default_ancestor), tree, shift)
                sir = sir + shift
                sor = sor + shift
            sil += vil.mod
            sir += vir.mod
            sol += vol.mod
            sor += vor.mod
        if vil.right() and not vor.right():
            vor.thread = vil.right()
            vor.mod += sil - sor
        else:
            if vir.left() and not vol.left():
                vol.thread = vir.left()
                vol.mod += sir - sol
            default_ancestor = tree
    return default_ancestor


def move_subtree(walk_l, walk_r, shift):
    """
    Determine possible shifts required to accomodate new node, but don't
    perform the shifts yet.
    """
    subtrees = walk_r.number - walk_l.number
    # print wl.tree, "is conflicted with", wr.tree, 'moving',
    # subtrees, 'shift', shift
    # print wl, wr, wr.number, wl.number, shift, subtrees, shift/subtrees
    walk_r.change -= shift / subtrees
    walk_r.shift += shift
    walk_l.change += shift / subtrees
    walk_r.coord_y += shift
    walk_r.mod += shift
    walk_r.height = max(walk_r.height, walk_r.coord_y)


def execute_shifts(tree):
    """
    Shift a tree, and it's subtrees, to allow for the placement of a
    new tree.
    """
    shift = change = 0
    for child in tree.children[::-1]:
        # print "shift:", child, shift, child.change
        child.coord_y += shift
        child.mod += shift
        change += child.change
        shift += child.shift + change
        child.height = max(child.height, child.coord_y)
        tree.height = max(tree.height, child.height)


def ancestor(vil, tree, default_ancestor):
    """
    The relevant text is at the bottom of page 7 of
    Improving Walker's Algorithm to Run in Linear Time" by Buchheim et al
    """
    if vil.ancestor in tree.parent.children:
        return vil.ancestor

    return default_ancestor


def second_walk(tree, modifier=0, h_separation=0, width=0, min_x=None):
    """
    Note that some of this code is modified to orientate the root node 'west'
    instead of 'north' in the Bushheim algorithms.
    """
    tree.coord_y += modifier
    tree.coord_x += width

    if min_x is None or tree.coord_x < min_x:
        min_x = tree.coord_x

    for child in tree.children:
        min_x = second_walk(
            child, modifier + tree.mod, h_separation, width + h_separation, min_x
        )
        tree.width = max(tree.width, child.width)
        tree.height = max(tree.height, child.height)

    tree.width = max(tree.width, tree.coord_x)
    tree.height = max(tree.height, tree.coord_y)
    return min_x



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010  Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import MatchesFilterBase


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Citation
from ....db import Database


# -------------------------------------------------------------------------
#
# MatchesFilter
#
# -------------------------------------------------------------------------
class MatchesSourceFilter(MatchesFilterBase):
    """
    Rule that checks against another filter.
    """

    labels = [_("Source filter name:")]
    name = _("Citations with source matching the <source filter>")
    description = _(
        "Matches citations with sources that match the " "specified source filter name"
    )
    category = _("General filters")

    # we want to have this filter show source filters
    namespace = "Source"

    def prepare(self, db: Database, user):
        MatchesFilterBase.prepare(self, db, user)
        self.MRF_filt = self.find_filter()

    def apply_to_one(self, db: Database, object: Citation) -> bool:
        if self.MRF_filt is None:
            return False

        source_handle = object.source_handle
        source = db.get_source_from_handle(source_handle)
        return self.MRF_filt.apply_to_one(db, source)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010        Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Tag editing module for Gramps.
"""
# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..managedwindow import ManagedWindow
from gramps.gen.const import URL_MANUAL_PAGE
from ..display import display_help
from ..listmodel import ListModel, TOGGLE

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Filters" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Tag_selection_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditTagList
#
# -------------------------------------------------------------------------
class EditTagList(ManagedWindow):
    """
    Dialog to allow the user to edit a list of tags.
    """

    def __init__(self, tag_list, full_list, uistate, track):
        """
        Initiate and display the dialog.
        """
        ManagedWindow.__init__(self, uistate, track, self, modal=True)
        # the self.window.run() below makes Gtk make it modal, so any change
        # to the previous line's "modal" would require that line to be changed

        self.namemodel = None
        top = self._create_dialog()
        self.set_window(top, None, _("Tag selection"))
        self.setup_configs("interface.edittaglist", 360, 400)

        for tag in full_list:
            self.namemodel.add([tag[0], tag in tag_list, tag[1]])
        self.namemodel.connect_model()

        # The dialog is modal.  We don't want to have several open dialogs of
        # this type, since then the user will loose track of which is which.
        self.return_list = None
        self.show()

        while True:
            # the self.window.run() makes Gtk make it modal, so any change to
            # that line means the ManagedWindow.__init__ must be changed also
            response = self.window.run()
            if response == Gtk.ResponseType.HELP:
                display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)
            elif response == Gtk.ResponseType.DELETE_EVENT:
                break
            else:
                if response == Gtk.ResponseType.OK:
                    self.return_list = [
                        (row[0], row[2]) for row in self.namemodel.model if row[1]
                    ]
                self.close()
                break

    def _create_dialog(self):
        """
        Create a dialog box to select tags.
        """
        # pylint: disable-msg=E1101
        top = Gtk.Dialog(transient_for=self.uistate.window)
        top.vbox.set_spacing(5)

        columns = [
            ("", -1, 300),
            (" ", -1, 25, TOGGLE, True, None),
            (_("Tag"), -1, 300),
        ]
        view = Gtk.TreeView()
        self.namemodel = ListModel(view, columns)

        slist = Gtk.ScrolledWindow()
        slist.add(view)
        slist.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        top.vbox.pack_start(slist, 1, 1, 5)

        top.add_button(_("_Help"), Gtk.ResponseType.HELP)
        top.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        top.add_button(_("_OK"), Gtk.ResponseType.OK)
        return top

    def build_menu_names(self, obj):  # meaningless while it's modal
        """
        Define the menu entry for the ManagedWindows.
        """
        return (_("Tag selection"), None)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007 Donald N. Allingham
# Copyright (C) 2011      Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Basic Primary Object class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from abc import abstractmethod

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .citationbase import CitationBase
from .mediabase import MediaBase
from .notebase import NoteBase
from .privacybase import PrivacyBase
from .tableobj import TableObject
from .tagbase import TagBase


# -------------------------------------------------------------------------
#
# BasicPrimaryObject
#
# -------------------------------------------------------------------------
class BasicPrimaryObject(TableObject, PrivacyBase, TagBase):
    """
    The BasicPrimaryObject is the base class for :class:`~.note.Note` objects.

    It is also the base class for the :class:`PrimaryObject` class.

    The :class:`PrimaryObject` is the base class for all other primary objects
    in the database. Primary objects are the core objects in the database.
    Each object has a database handle and a Gramps ID value. The database
    handle is used as the record number for the database, and the Gramps
    ID is the user visible version.
    """

    def __init__(self, source=None):
        """
        Initialize a PrimaryObject.

        If source is None, both the ID and handle are assigned as empty
        strings. If source is not None, then object is initialized from values
        of the source object.

        :param source: Object used to initialize the new object
        :type source: PrimaryObject
        """
        TableObject.__init__(self, source)
        PrivacyBase.__init__(self, source)
        TagBase.__init__(self)
        if source:
            self.gramps_id = source.gramps_id
        else:
            self.gramps_id = None

    @abstractmethod
    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """

    @abstractmethod
    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """

    def set_gramps_id(self, gramps_id):
        """
        Set the Gramps ID for the primary object.

        :param gramps_id: Gramps ID
        :type gramps_id: str
        """
        self.gramps_id = gramps_id

    def get_gramps_id(self):
        """
        Return the Gramps ID for the primary object.

        :returns: Gramps ID associated with the object
        :rtype: str
        """
        return self.gramps_id

    def has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str

        :returns:
          Returns whether the object has reference to this handle of
          this object type.

        :rtype: bool
        """
        return False

    def remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """

    def replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """

    def has_citation_reference(self, handle):
        """
        Indicate if the object has a citation references.

        In the base class, no such references exist. Derived classes should
        override this if they provide citation references.
        """
        return False

    def has_media_reference(self, handle):
        """
        Indicate if the object has a media references.

        In the base class, no such references exist. Derived classes should
        override this if they provide media references.
        """
        return False

    def remove_citation_references(self, handle_list):
        """
        Remove the specified citation references from the object.

        In the base class no such references exist. Derived classes should
        override this if they provide citation references.
        """

    def remove_media_references(self, handle_list):
        """
        Remove the specified media references from the object.

        In the base class no such references exist. Derived classes should
        override this if they provide media references.
        """

    def remove_note_references(self, handle_list):
        """
        Remove the specified note references from the object.

        In the base class no such references exist. Derived classes should
        override this if they provide note references.
        """

    def replace_citation_references(self, old_handle, new_handle):
        """
        Replace all references to the old citation handle with those to the new
        citation handle.
        """

    def replace_media_references(self, old_handle, new_handle):
        """
        Replace all references to the old media handle with those to the new
        media handle.
        """


# -------------------------------------------------------------------------
#
# Primary Object class
#
# -------------------------------------------------------------------------
class PrimaryObject(BasicPrimaryObject):
    """
    The PrimaryObject is the base class for all primary objects in the
    database.

    Primary objects are the core objects in the database.
    Each object has a database handle and a Gramps ID value. The database
    handle is used as the record number for the database, and the Gramps
    ID is the user visible version.
    """

    def __init__(self, source=None):
        """
        Initialize a PrimaryObject.

        If source is None, both the ID and handle are assigned as empty
        strings. If source is not None, then object is initialized from values
        of the source object.

        :param source: Object used to initialize the new object
        :type source: PrimaryObject
        """
        BasicPrimaryObject.__init__(self, source)

    @abstractmethod
    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """

    @abstractmethod
    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """

    def has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle
                  of this object type.
        :rtype: bool
        """
        if classname == "Citation" and isinstance(self, CitationBase):
            return self.has_citation_reference(handle)
        if classname == "Media" and isinstance(self, MediaBase):
            return self.has_media_reference(handle)
        return self._has_handle_reference(classname, handle)

    def remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """
        if classname == "Citation" and isinstance(self, CitationBase):
            self.remove_citation_references(handle_list)
        elif classname == "Media" and isinstance(self, MediaBase):
            self.remove_media_references(handle_list)
        elif classname == "Note" and isinstance(self, NoteBase):
            self.remove_note_references(handle_list)
        else:
            self._remove_handle_references(classname, handle_list)

    def replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Citation" and isinstance(self, CitationBase):
            self.replace_citation_references(old_handle, new_handle)
        elif classname == "Media" and isinstance(self, MediaBase):
            self.replace_media_references(old_handle, new_handle)
        else:
            self._replace_handle_reference(classname, old_handle, new_handle)

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the handle is referenced by the object.
        """
        return False

    def _remove_handle_references(self, classname, handle_list):
        """
        Remove the handle references from the object.
        """

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace the handle reference with the new reference.
        """



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.datehandler import format_time
from .flatbasemodel import FlatBaseModel
from gramps.gen.const import GRAMPS_LOCALE as glocale


# -------------------------------------------------------------------------
#
# SourceModel
#
# -------------------------------------------------------------------------
class SourceModel(FlatBaseModel):
    def __init__(
        self,
        db,
        uistate,
        scol=0,
        order=Gtk.SortType.ASCENDING,
        search=None,
        skip=set(),
        sort_map=None,
    ):
        self.map = db.get_raw_source_data
        self.gen_cursor = db.get_source_cursor
        self.fmap = [
            self.column_title,
            self.column_id,
            self.column_author,
            self.column_abbrev,
            self.column_pubinfo,
            self.column_private,
            self.column_tags,
            self.column_change,
            self.column_tag_color,
        ]
        self.smap = [
            self.column_title,
            self.column_id,
            self.column_author,
            self.column_abbrev,
            self.column_pubinfo,
            self.column_private,
            self.column_tags,
            self.sort_change,
            self.column_tag_color,
        ]
        FlatBaseModel.__init__(
            self, db, uistate, scol, order, search=search, skip=skip, sort_map=sort_map
        )

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        self.db = None
        self.gen_cursor = None
        self.map = None
        self.fmap = None
        self.smap = None
        FlatBaseModel.destroy(self)

    def color_column(self):
        """
        Return the color column.
        """
        # This is for model.get_value() arg
        return 8

    def on_get_n_columns(self):
        return len(self.fmap) + 1

    def column_title(self, data):
        return data.title.replace("\n", " ")

    def column_author(self, data):
        return data.author

    def column_abbrev(self, data):
        return data.abbrev

    def column_id(self, data):
        return data.gramps_id

    def column_pubinfo(self, data):
        return data.pubinfo

    def column_private(self, data):
        if data.private:
            return "gramps-lock"
        else:
            # There is a problem returning None here.
            return ""

    def column_change(self, data):
        return format_time(data.change)

    def sort_change(self, data):
        return "%012x" % data.change

    def get_tag_name(self, tag_handle):
        """
        Return the tag name from the given tag handle.
        """
        cached, value = self.get_cached_value(tag_handle, "TAG_NAME")
        if not cached:
            value = self.db.get_tag_from_handle(tag_handle).get_name()
            self.set_cached_value(tag_handle, "TAG_NAME", value)
        return value

    def column_tag_color(self, data):
        """
        Return the tag color.
        """
        tag_handle = data.handle
        cached, value = self.get_cached_value(tag_handle, "TAG_COLOR")
        if not cached:
            tag_color = ""
            tag_priority = None
            for handle in data.tag_list:
                tag = self.db.get_tag_from_handle(handle)
                if tag:
                    this_priority = tag.get_priority()
                    if tag_priority is None or this_priority < tag_priority:
                        tag_color = tag.get_color()
                        tag_priority = this_priority
            value = tag_color
            self.set_cached_value(tag_handle, "TAG_COLOR", value)
        return value

    def column_tags(self, data):
        """
        Return the sorted list of tags.
        """
        tag_list = list(map(self.get_tag_name, data.tag_list))
        # TODO for Arabic, should the next line's comma be translated?
        return ", ".join(sorted(tag_list, key=glocale.sort_key))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008  Jerome Rapinat
# Copyright (C) 2008  Benny Malengier
# Copyright (C) 2011        Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._hasnotetypebase import HasNoteTypeBase


# -------------------------------------------------------------------------
# "Sources having notes"
# -------------------------------------------------------------------------
class HasNoteType(HasNoteTypeBase):
    """Sources with a note of the specified type"""

    name = _("Sources with a note of type <type>")
    description = _("Matches sources with a note of the specified type")
    category = _("Note filters")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017  Paul Culley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""unittest for styledtext"""

import unittest
from copy import deepcopy

from ..styledtext import StyledText
from ..styledtexttag import StyledTextTag
from ..styledtexttagtype import StyledTextTagType


class Test1(unittest.TestCase):
    T1 = StyledTextTag(StyledTextTagType(1), "v1", [(0, 2), (2, 4), (4, 6)])
    T2 = StyledTextTag(StyledTextTagType(2), "v2", [(1, 3), (3, 5), (0, 7)])
    T3 = StyledTextTag(StyledTextTagType(0), "v3", [(0, 1)])
    T4 = StyledTextTag(StyledTextTagType(2), "v2", [(8, 10), (10, 12), (7, 14)])
    T5 = StyledTextTag(StyledTextTagType(2), "v2", [(19, 21), (21, 23), (18, 25)])

    A = StyledText("123X456", [T1])
    B = StyledText("abcXdef", [T2])

    C = StyledText("\n")

    S = "cleartext"

    # some basic tests
    # because the StyledText.__eq__ method doesn't work very well (tags don't
    # compare when they are equivalent, but not equal) we have to use
    # serialize for comparisons.
    def test_join(self):
        C = self.C.join([self.A, self.S, deepcopy(self.B)])
        _C = StyledText("123X456\ncleartext\nabcXdef", [self.T1, self.T5])
        self.assertEqual(C.serialize(), _C.serialize())

    def test_split(self):
        C = self.C.join([self.A, self.S, deepcopy(self.B)])
        L = C.split()
        _L = [self.A, self.S, self.B]
        self.assertEqual(L[0].serialize(), self.A.serialize())
        self.assertEqual(str(L[1]), self.S)
        self.assertEqual(L[2].serialize(), self.B.serialize())

    def test_replace(self):
        C = self.C.join([self.A, self.S, deepcopy(self.B)])
        C = C.replace("X", StyledText("_", [self.T3]))
        _C = (
            "123_456\ncleartext\nabc_def",
            [
                ((0, ""), "v3", [(3, 4)]),
                ((0, ""), "v3", [(21, 22)]),
                ((1, ""), "v1", [(0, 2), (2, 3)]),
                ((1, ""), "v1", [(4, 6)]),
                ((2, ""), "v2", [(19, 21), (18, 21)]),
                ((2, ""), "v2", [(22, 23), (22, 25)]),
            ],
        )
        self.assertEqual(C.serialize(), _C)

    def test_add(self):
        A = deepcopy(self.A) + deepcopy(self.B)
        _A = StyledText("123X456abcXdef", [self.T1, self.T4])
        self.assertEqual(A.serialize(), _A.serialize())


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# gtk
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ... import widgets
from gramps.gen.lib import Date, Event, EventType
from gramps.gen.datehandler import displayer
from .. import build_filter_model
from . import SidebarFilter
from gramps.gen.filters import GenericFilter, rules
from gramps.gen.filters.rules.person import (
    RegExpName,
    RegExpIdOf,
    IsMale,
    IsFemale,
    HasUnknownGender,
    HasEvent,
    HasTag,
    HasBirth,
    HasDeath,
    HasNoteRegexp,
    MatchesFilter,
    HasOtherGender,
)


def extract_text(entry_widget):
    """
    Extract the text from the entry widget, strips off any extra spaces,
    and converts the string to unicode.

    For some strange reason a gtk bug prevents the extracted string from being
    of type unicode.

    """
    return str(entry_widget.get_text().strip())


# -------------------------------------------------------------------------
#
# PersonSidebarFilter class
#
# -------------------------------------------------------------------------
class PersonSidebarFilter(SidebarFilter):
    def __init__(self, dbstate, uistate, clicked):
        self.clicked_func = clicked
        self.filter_name = widgets.BasicEntry()
        self.filter_id = widgets.BasicEntry()
        self.filter_birth = widgets.DateEntry(uistate, [])
        self.filter_birth_place = widgets.BasicEntry()
        self.filter_death = widgets.DateEntry(uistate, [])
        self.filter_death_place = widgets.BasicEntry()
        self.filter_event = Event()
        self.filter_event.set_type((EventType.CUSTOM, ""))
        self.etype = Gtk.ComboBox(has_entry=True)
        if dbstate.is_open():
            self.custom_types = dbstate.db.get_event_types()
        else:
            self.custom_types = []
        self.event_menu = widgets.MonitoredDataType(
            self.etype,
            self.filter_event.set_type,
            self.filter_event.get_type,
            custom_values=self.custom_types,
        )

        self.filter_note = widgets.BasicEntry()
        self.filter_gender = Gtk.ComboBoxText()
        list(
            map(
                self.filter_gender.append_text,
                [
                    _("any"),
                    _("male"),
                    _("female"),
                    _("other", "gender"),
                    _("unknown", "gender"),
                ],
            )
        )
        self.filter_gender.set_active(0)

        self.filter_regex = Gtk.CheckButton(label=_("Use regular expressions"))
        self.sensitive_regex = Gtk.CheckButton(label=_("Case sensitive"))

        self.tag = Gtk.ComboBox()
        self.generic = Gtk.ComboBox()

        SidebarFilter.__init__(self, dbstate, uistate, "Person")

    def create_widget(self):
        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.generic.pack_start(cell, True)
        self.generic.add_attribute(cell, "text", 0)
        self.on_filters_changed("Person")

        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.tag.pack_start(cell, True)
        self.tag.add_attribute(cell, "text", 0)

        self.etype.get_child().set_width_chars(5)

        exdate1 = Date()
        exdate2 = Date()
        exdate1.set(
            Date.QUAL_NONE,
            Date.MOD_RANGE,
            Date.CAL_GREGORIAN,
            (0, 0, 1800, False, 0, 0, 1900, False),
        )
        exdate2.set(
            Date.QUAL_NONE, Date.MOD_BEFORE, Date.CAL_GREGORIAN, (0, 0, 1850, False)
        )

        msg1 = displayer.display(exdate1)
        msg2 = displayer.display(exdate2)

        self.add_text_entry(_("Name"), self.filter_name)
        self.add_text_entry(_("ID"), self.filter_id)
        self.add_entry(_("Gender"), self.filter_gender)
        self.add_text_entry(
            _("Birth date"),
            self.filter_birth,
            _('example: "%(msg1)s" or "%(msg2)s"') % {"msg1": msg1, "msg2": msg2},
        )
        self.add_text_entry(_("Birth place"), self.filter_birth_place)
        self.add_text_entry(
            _("Death date"),
            self.filter_death,
            _('example: "%(msg1)s" or "%(msg2)s"') % {"msg1": msg1, "msg2": msg2},
        )
        self.add_text_entry(_("Death place"), self.filter_death_place)
        self.add_entry(_("Event"), self.etype)
        self.add_text_entry(_("Note"), self.filter_note)
        self.add_entry(_("Tag"), self.tag)
        self.add_filter_entry(_("Custom filter"), self.generic)
        self.add_regex_entry(self.filter_regex)
        self.add_regex_case(self.sensitive_regex)

    def clear(self, obj):
        self.filter_name.set_text("")
        self.filter_id.set_text("")
        self.filter_birth.set_text("")
        self.filter_birth_place.set_text("")
        self.filter_death.set_text("")
        self.filter_death_place.set_text("")
        self.filter_note.set_text("")
        self.filter_gender.set_active(0)
        self.etype.get_child().set_text("")
        self.tag.set_active(0)
        self.generic.set_active(0)

    def get_filter(self):
        """
        Extracts the text strings from the sidebar, and uses them to build up
        a new filter.
        """

        # extract text values from the entry widgets
        name = extract_text(self.filter_name)
        gid = extract_text(self.filter_id)
        birth = extract_text(self.filter_birth)
        birth_place = extract_text(self.filter_birth_place)
        death = extract_text(self.filter_death)
        death_place = extract_text(self.filter_death_place)
        note = extract_text(self.filter_note)

        # extract remaining data from the menus
        etype = self.filter_event.get_type().xml_str()
        gender = self.filter_gender.get_active()
        regex = self.filter_regex.get_active()
        usecase = self.sensitive_regex.get_active()
        tag = self.tag.get_active() > 0
        generic = self.generic.get_active() > 0

        # check to see if the filter is empty. If it is empty, then
        # we don't build a filter

        empty = not (
            name
            or gid
            or birth
            or birth_place
            or death
            or death_place
            or etype
            or note
            or gender
            or regex
            or tag
            or generic
        )
        if empty:
            generic_filter = None
        else:
            # build a GenericFilter
            generic_filter = GenericFilter()

            # if the name is not empty, split the name in multiple part if
            # we don't use regexp. if the regexp is used, don't split the
            # field
            if name:
                if not regex:
                    name_parts = name.split(sep=" ")
                    for name_part in name_parts:
                        rule = RegExpName(
                            [name_part], use_regex=regex, use_case=usecase
                        )
                        generic_filter.add_rule(rule)
                else:
                    rule = RegExpName([name], use_regex=regex, use_case=usecase)
                    generic_filter.add_rule(rule)

            # if the id is not empty, choose either the regular expression
            # version or the normal text match
            if gid:
                rule = RegExpIdOf([gid], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            # check the gender, and select the right rule based on gender
            if gender > 0:
                if gender == 1:
                    generic_filter.add_rule(IsMale([]))
                elif gender == 2:
                    generic_filter.add_rule(IsFemale([]))
                elif gender == 3:
                    generic_filter.add_rule(HasOtherGender([]))
                else:
                    generic_filter.add_rule(HasUnknownGender([]))

            # Build an event filter if needed
            if etype:
                rule = HasEvent(
                    [etype, "", "", "", "", "1"], use_regex=regex, use_case=usecase
                )
                generic_filter.add_rule(rule)

            # Build birth event filter if needed
            # Arguments for the HasBirth filter are Date, Place, and Description
            # Since the value we extracted to the "birth" variable is the
            # request date, we pass it as the first argument
            if birth or birth_place:
                rule = HasBirth(
                    [birth, birth_place, ""], use_regex=regex, use_case=usecase
                )
                generic_filter.add_rule(rule)

            # Build death event filter if needed
            if death or death_place:
                rule = HasDeath(
                    [death, death_place, ""], use_regex=regex, use_case=usecase
                )
                generic_filter.add_rule(rule)

            # Build note filter if needed
            if note:
                rule = HasNoteRegexp([note], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            # check the Tag
            if tag:
                model = self.tag.get_model()
                node = self.tag.get_active_iter()
                attr = model.get_value(node, 0)
                rule = HasTag([attr])
                generic_filter.add_rule(rule)

        if self.generic.get_active() != 0:
            model = self.generic.get_model()
            node = self.generic.get_active_iter()
            obj = str(model.get_value(node, 0))
            rule = MatchesFilter([obj])
            generic_filter.add_rule(rule)

        return generic_filter

    def on_filters_changed(self, name_space):
        if name_space == "Person":
            all_filter = GenericFilter()
            all_filter.set_name(_("None"))
            all_filter.add_rule(rules.person.Everyone([]))
            self.generic.set_model(build_filter_model("Person", [all_filter]))
            self.generic.set_active(0)

    def on_tags_changed(self, tag_list):
        """
        Update the list of tags in the tag filter.
        """
        model = Gtk.ListStore(str)
        model.append(("",))
        for tag_name in tag_list:
            model.append((tag_name,))
        self.tag.set_model(model)
        self.tag.set_active(0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..managedwindow import ManagedWindow
from ..display import display_help
from gramps.gen.config import config
from ..dbguielement import DbGUIElement


class EditSecondary(ManagedWindow, DbGUIElement):
    def __init__(self, state, uistate, track, obj, callback=None):
        """Create an edit window.  Associates a person with the window."""

        self.obj = obj
        self.old_obj = state.db.serializer.object_to_data(obj)
        self.dbstate = state
        self.uistate = uistate
        self.db = state.db
        self.callback = callback
        self.__tabs = []

        ManagedWindow.__init__(self, uistate, track, obj)
        DbGUIElement.__init__(self, self.db)

        self._local_init()
        self._set_size()

        self._create_tabbed_pages()
        self._setup_fields()
        self._connect_signals()
        self.show()
        self._post_init()

    def _local_init(self):
        """
        Derived class should do any pre-window initalization in this task.
        """
        pass

    def _post_init(self):
        """
        Derived class should do any post-window initalization in this task.
        """
        pass

    def _connect_signals(self):
        pass

    def _setup_fields(self):
        pass

    def _create_tabbed_pages(self):
        pass

    def build_window_key(self, obj):
        return id(obj)

    def _setup_notebook_tabs(self, notebook):
        for child in notebook.get_children():
            label = notebook.get_tab_label(child)
            page_no = notebook.page_num(child)
            label.drag_dest_set(0, [], 0)
            label.connect("drag_motion", self._switch_page_on_dnd, notebook, page_no)
            child.set_parent_notebook(notebook)
        notebook.connect("key-press-event", self.key_pressed, notebook)

    def key_pressed(self, obj, event, notebook):
        """
        Handles the key being pressed on the notebook, pass to key press of
        current page.
        """
        pag = notebook.get_current_page()
        if not pag == -1:
            notebook.get_nth_page(pag).key_pressed(obj, event)

    def _switch_page_on_dnd(self, widget, context, x, y, time, notebook, page_no):
        if notebook.get_current_page() != page_no:
            notebook.set_current_page(page_no)

    def _add_tab(self, notebook, page):
        self.__tabs.append(page)
        notebook.insert_page(page, page.get_tab_widget(), -1)
        page.label.set_use_underline(True)
        return page

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        for tab in self.__tabs:
            if hasattr(tab, "_cleanup_on_exit"):
                tab._cleanup_on_exit()
        self.__tabs = None
        self.dbstate = None
        self.uistate = None
        self.obj = None
        self.db = None
        self.callback = None
        self.callman.database = None
        self.callman = None

    def define_ok_button(self, button, function):
        button.connect("clicked", function)
        button.set_sensitive(not self.db.readonly)

    def define_cancel_button(self, button):
        button.connect("clicked", self.canceledits)

    def define_help_button(self, button, webpage="", section=""):
        button.connect("clicked", lambda x: display_help(webpage, section))

    def canceledits(self, *obj):
        """
        Undo the edits that happened on this secondary object
        """
        self.obj = self.db.serializer.data_to_object(self.old_obj)
        self.close(obj)

    def close(self, *obj):
        self._cleanup_db_connects()
        self._cleanup_connects()
        ManagedWindow.close(self)
        self._cleanup_on_exit()

    def _cleanup_db_connects(self):
        """
        All connects that happened to signals of the db must be removed on
        closed. This implies two things:
        1. The connects on the main view must be disconnected
        2. Connects done in subelements must be disconnected
        """
        # cleanup callbackmanager of this editor
        self._cleanup_callbacks()
        for tab in [tab for tab in self.__tabs if hasattr(tab, "callman")]:
            tab._cleanup_callbacks()

    def _cleanup_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface
        """
        self._cleanup_local_connects()
        for tab in [
            tab for tab in self.__tabs if hasattr(tab, "_cleanup_local_connects")
        ]:
            tab._cleanup_local_connects()

    def _cleanup_local_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface. This methods cleans connects
        of the main interface, not of the displaytabs.
        """
        pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011       Robert Cheramy
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# RelatedWith
#
# -------------------------------------------------------------------------
class IsRelatedWith(Rule):
    """Rule that checks if a person is related to a specified person"""

    labels = [_("ID:")]
    name = _("People related to <Person>")
    category = _("Relationship filters")
    description = _("Matches people related to a specified person")

    def prepare(self, db: Database, user):
        """prepare so the rule can be executed efficiently
        we build the list of people related to <person> here,
        so that apply is only a check into this list
        """
        self.db = db

        self.selected_handles: Set[str] = set()
        self.add_relative(db.get_person_from_gramps_id(self.list[0]))

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def add_relative(self, start: Person | None):
        """Non-recursive function that scans relatives and add them to self.selected_handles"""
        if not start:
            return

        queue: List[Person] = [start]

        while queue:
            person = queue.pop()
            # Add the relative to the list
            if person is None or (person.handle in self.selected_handles):
                continue
            self.selected_handles.add(person.handle)

            for family_handle in person.parent_family_list:
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    # Check Parents
                    for parent_handle in (
                        family.father_handle,
                        family.mother_handle,
                    ):
                        if parent_handle:
                            queue.append(self.db.get_person_from_handle(parent_handle))
                    # Check Sibilings
                    for child_ref in family.child_ref_list:
                        queue.append(self.db.get_person_from_handle(child_ref.ref))

            for family_handle in person.family_list:
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    # Check Spouse
                    for parent_handle in (
                        family.father_handle,
                        family.mother_handle,
                    ):
                        if parent_handle:
                            queue.append(self.db.get_person_from_handle(parent_handle))
                    # Check Children
                    for child_ref in family.child_ref_list:
                        queue.append(self.db.get_person_from_handle(child_ref.ref))

        return



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from .selectorexceptions import SelectorException


def SelectorFactory(classname):
    if classname == "Person":
        from .selectperson import SelectPerson

        cls = SelectPerson
    elif classname == "Family":
        from .selectfamily import SelectFamily

        cls = SelectFamily
    elif classname == "Event":
        from .selectevent import SelectEvent

        cls = SelectEvent
    elif classname == "Place":
        from .selectplace import SelectPlace

        cls = SelectPlace
    elif classname == "Source":
        from .selectsource import SelectSource

        cls = SelectSource
    elif classname == "Citation":
        from .selectcitation import SelectCitation

        cls = SelectCitation
    elif classname == "Media":
        from .selectobject import SelectObject

        cls = SelectObject
    elif classname == "Repository":
        from .selectrepository import SelectRepository

        cls = SelectRepository
    elif classname == "Note":
        from .selectnote import SelectNote

        cls = SelectNote
    else:
        raise SelectorException(
            "Attempt to create unknown "
            "selector class: "
            "classname = %s" % (str(classname),)
        )

    return cls



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2009       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os
from gi.repository import Gtk
from gi.repository import GObject

from gramps.gen.constfunc import get_curr_dir
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class FileEntry(Gtk.Box):
    """A widget that allows the user to select a file from the file system"""

    def __init__(self, defname, title, parent=None):
        Gtk.Box.__init__(self)

        self.title = title
        self.parent = parent
        self.dir = False
        self.__base_path = ""
        self.__file_name = ""
        self.entry = Gtk.Entry()
        self.entry.set_text(defname)
        self.set_filename(defname)
        self.set_spacing(6)
        self.set_homogeneous(False)
        self.button = Gtk.Button()
        image = Gtk.Image()
        image.set_from_icon_name("document-open", Gtk.IconSize.BUTTON)
        self.button.add(image)
        self.button.connect("clicked", self.__select_file)
        self.pack_start(self.entry, True, True, 0)
        self.pack_end(self.button, False, False, 0)

    def __select_file(self, obj):
        """Call back function to handle the open button press"""
        if self.dir:
            my_action = Gtk.FileChooserAction.SELECT_FOLDER
        else:
            my_action = Gtk.FileChooserAction.SAVE

        dialog = Gtk.FileChooserDialog(
            title=self.title, transient_for=self.parent, action=my_action
        )
        dialog.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("_Open"), Gtk.ResponseType.OK
        )

        name = os.path.basename(self.entry.get_text())
        if self.dir:
            if os.path.isdir(name):
                dialog.set_current_name(name)
            elif os.path.isdir(os.path.basename(name)):
                dialog.set_current_name(os.path.basename(name))
        else:
            dialog.set_current_name(name)
        dialog.set_current_folder(self.__base_path)
        dialog.present()
        status = dialog.run()
        if status == Gtk.ResponseType.OK:
            self.set_filename(dialog.get_filename())
        dialog.destroy()

    def set_filename(self, path):
        """Set the currently selected dialog."""
        if not path:
            return
        if os.path.dirname(path):
            self.__base_path = os.path.dirname(path)
            self.__file_name = os.path.basename(path)
        else:
            self.__base_path = get_curr_dir()
            self.__file_name = path
        self.entry.set_text(os.path.join(self.__base_path, self.__file_name))

    def get_full_path(self, val):
        """Get the full path of the currently selected file."""
        return self.entry.get_text()

    def set_directory_entry(self, opt):
        """
        Configure the FileEntry to either select a directory or a file.
        Set it to True to select a directory.
        Set it to False to select a file.
        """
        self.dir = opt



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2008  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2008       Robert Cheramy <robert@cheramy.net>
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Export to Gramps package"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import time
import shutil
import os
import tarfile
from io import StringIO, BytesIO
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".WritePkg")

# -------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.plug.export import WriterOptionBox
from gramps.plugins.export.exportxml import XmlWriter
from gramps.gen.utils.file import media_path_full
from gramps.gen.constfunc import win


# -------------------------------------------------------------------------
#
# writeData
#
# -------------------------------------------------------------------------
def writeData(database, filename, user, option_box=None):
    # Rename file, if it exists already, with <filename>.bak
    # as it it for normal XML export.

    if os.path.isfile(filename):
        try:
            shutil.copyfile(filename, filename + ".bak")
            shutil.copystat(filename, filename + ".bak")
        except:
            pass

    if option_box:
        option_box.parse_options()
        database = option_box.get_filtered_database(database)

    writer = PackageWriter(database, filename, user)
    return writer.export()


def fix_mtime(tarinfo):
    """this fixes a bug in the python tarfile GNU_FORMAT where if mtime has
    a fractional component, it fails."""
    tarinfo.mtime = int(tarinfo.mtime)
    return tarinfo


# -------------------------------------------------------------------------
#
# PackageWriter
#
# -------------------------------------------------------------------------
class PackageWriter:
    def __init__(self, database, filename, user):
        self.db = database
        self.user = user
        self.filename = filename

    def export(self):
        #        missmedia_action = 0
        # --------------------------------------------------------------
        # def remove_clicked():
        #     # File is lost => remove all references and the object itself
        #     for p_id in self.db.iter_family_handles():
        #         p = self.db.get_family_from_handle(p_id)
        #         nl = p.get_media_list()
        #         for o in nl:
        #             if o.get_reference_handle() == m_id:
        #                 nl.remove(o)
        #         p.set_media_list(nl)
        #         self.db.commit_family(p,None)
        #     for key in self.db.iter_person_handles():
        #         p = self.db.get_person_from_handle(key)
        #         nl = p.get_media_list()
        #         for o in nl:
        #             if o.get_reference_handle() == m_id:
        #                 nl.remove(o)
        #         p.set_media_list(nl)
        #         self.db.commit_person(p,None)
        #     for key in self.db.get_source_handles():
        #         p = self.db.get_source_from_handle(key)
        #         nl = p.get_media_list()
        #         for o in nl:
        #             if o.get_reference_handle() == m_id:
        #                 nl.remove(o)
        #         p.set_media_list(nl)
        #         self.db.commit_source(p,None)
        #     for key in self.db.get_place_handles():
        #         p = self.db.get_place_from_handle(key)
        #         nl = p.get_media_list()
        #         for o in nl:
        #             if o.get_reference_handle() == m_id:
        #                 nl.remove(o)
        #         p.set_media_list(nl)
        #         self.db.commit_place(p,None)
        #     for key in self.db.get_event_handles():
        #         p = self.db.get_event_from_handle(key)
        #         nl = p.get_media_list()
        #         for o in nl:
        #             if o.get_reference_handle() == m_id:
        #                 nl.remove(o)
        #         p.set_media_list(nl)
        #         self.db.commit_event(p,None)
        #     self.db.remove_media(m_id,None)

        # def leave_clicked():
        #     # File is lost => do nothing, leave as is
        #     pass

        # def select_clicked():
        #     # File is lost => select a file to replace the lost one
        #     def fs_close_window(obj):
        #         pass

        #     def fs_ok_clicked(obj):
        #         name = fs_top.get_filename()
        #         if os.path.isfile(name):
        #             archive.add(name)

        #     fs_top = gtk.FileChooserDialog("%s - GRAMPS" % _("Select file"),
        #                 buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
        #                          gtk.STOCK_OK, Gtk.ResponseType.OK)
        #                 )
        #     response = fs_top.run()
        #     if response == Gtk.ResponseType.OK:
        #         fs_ok_clicked(fs_top)
        #     elif response == gtk.RESPONSE_CANCEL:
        #         fs_close_window(fs_top)

        #     fs_top.destroy()
        # ---------------------------------------------------------------

        try:
            with tarfile.open(self.filename, "w:gz") as archive:
                # Write media files first, since the database may be modified
                # during the process (i.e. when removing object)
                handles = self.db.get_media_handles(sort_handles=True)
                for indx, m_id in enumerate(handles):
                    self.user.callback(indx * 100 / len(handles))
                    mobject = self.db.get_media_from_handle(m_id)
                    filename = media_path_full(self.db, mobject.get_path())
                    archname = str(mobject.get_path())
                    if os.path.isfile(filename) and os.access(filename, os.R_OK):
                        archive.add(filename, archname, filter=fix_mtime)

                # Write XML now
                with BytesIO() as g:
                    gfile = XmlWriter(self.db, self.user, 2)
                    gfile.write_handle(g)
                    tarinfo = tarfile.TarInfo("data.gramps")
                    tarinfo.size = len(g.getvalue())
                    tarinfo.mtime = time.time()
                    if not win():
                        tarinfo.uid = os.getuid()
                        tarinfo.gid = os.getgid()
                    g.seek(0)
                    archive.addfile(tarinfo, g)

                return True
        except (EnvironmentError, OSError) as msg:
            log.warning(str(msg))
            self.user.notify_error(_("Failure writing %s") % self.filename, str(msg))
            return 0
